from __future__ import absolute_import, with_statement

import configparser
import getpass
import json
import logging
import logging.config
import os
import platform
import time

import keyring
import requests
import yaml
from pkg_resources import resource_filename
from requests import HTTPError

import NDATools
from NDATools import NDA_TOOLS_LOGGING_YML_FILE, Utils
from NDATools.Utils import exit_error, HttpErrorHandlingStrategy

logger = logging.getLogger(__name__)


class LoggingConfiguration:

    def __init__(self):
        pass

    @staticmethod
    def load_config(default_log_directory, verbose=False, log_dir=None):

        with open(NDA_TOOLS_LOGGING_YML_FILE, 'r') as stream:
            config = yaml.load(stream, Loader=yaml.FullLoader)
        if log_dir and os.path.exists(log_dir):
            log_file = os.path.join(log_dir, "debug_log_{}.txt").format(time.strftime("%Y%m%dT%H%M%S"))
        else:
            log_file = os.path.join(default_log_directory, "debug_log_{}.txt").format(time.strftime("%Y%m%dT%H%M%S"))
        config['handlers']['file']['filename'] = log_file
        if verbose:
            config['loggers']['NDATools']['level'] = 'DEBUG'
            config['handlers']['console']['formatter'] = 'detailed'
        logging.config.dictConfig(config)


class ClientConfiguration:
    SERVICE_NAME = 'nda-tools'

    def __init__(self, args):
        self.config = configparser.ConfigParser()
        logger.info('Using configuration file from {}'.format(NDATools.NDA_TOOLS_SETTINGS_CFG_FILE))
        self.config.read(NDATools.NDA_TOOLS_SETTINGS_CFG_FILE)
        self._check_and_fix_missing_options()
        self.validation_api = self.config.get("Endpoints", "validation")
        self.submission_package_api = self.config.get("Endpoints", "submission_package")
        self.submission_api = self.config.get("Endpoints", "submission")
        self.validationtool_api = self.config.get("Endpoints", "validationtool")
        self.package_creation_api = self.config.get("Endpoints", "package_creation")
        self.package_api = self.config.get("Endpoints", "package")
        self.datadictionary_api = self.config.get("Endpoints", "datadictionary")
        self.collection_api = self.config.get("Endpoints", "collection")
        self.user_api = self.config.get("Endpoints", "user")
        self.aws_access_key = self.config.get("User", "access_key")
        self.aws_secret_key = self.config.get("User", "secret_key")
        self.aws_session_token = self.config.get('User', 'session_token')
        self.username = self.config.get("User", "username").lower()

        # options that appear in both vtcmd and downloadcmd
        self.workerThreads = args.workerThreads

        if args.username:
            self.username = args.username
        elif self.username:
            logger.warning("-u/--username argument not provided. Using default value of '%s' which was saved in %s",
                           self.username, NDATools.NDA_TOOLS_SETTINGS_CFG_FILE)
        self.password = None

        is_vtcmd = 'collectionID' in args
        if is_vtcmd:
            self.aws_access_key = args.accessKey
            self.aws_secret_key = args.secretKey
            self.hideProgress = args.hideProgress
            self.force = True if args.force else False
            self.collection_id = args.collectionID
            self.directory_list = args.listDir
            self.manifest_path = args.manifestPath
            self.source_bucket = args.s3Bucket
            self.source_prefix = args.s3Prefix
            self.validation_timeout = args.validation_timeout
            self.title = args.title
            self.description = args.description
            self.scope = args.scope
            self.JSON = args.JSON
            self.skip_local_file_check = args.skipLocalAssocFileCheck
            self.replace_submission = args.replace_submission
        logger.info('proceeding as nda user: {}'.format(self.username))

    def _check_and_fix_missing_options(self):
        default_config = configparser.ConfigParser()
        default_file_path = resource_filename(__name__, 'clientscripts/config/settings.cfg')
        default_config.read(default_file_path)
        change_detected = False
        for section in default_config.sections():
            if section not in self.config.sections():
                logger.debug(f'adding {section} to settings.cfg')
                self.config.add_section(section)
                change_detected = True
            for option in default_config[section]:
                if option not in self.config[section]:
                    logger.debug('[{}][{}] is missing'.format(section, option))
                    self.config.set(section, option, default_config[section][option])
                    change_detected = True
        if change_detected:
            logger.debug(f'updating settings.cfg')
            with open(NDATools.NDA_TOOLS_SETTINGS_CFG_FILE, 'w') as configfile:
                self.config.write(configfile)
        else:
            logger.debug(f'settings.cfg is up to date')

    def read_user_credentials(self, auth_req=True):

        if auth_req:
            while True:
                try:
                    while not self.username:
                        self.username = str(input('Enter your NIMH Data Archives username:')).lower()
                        with open(NDATools.NDA_TOOLS_SETTINGS_CFG_FILE, 'w') as configfile:
                            self.config.set('User', 'username', self.username)
                            self.config.write(configfile)
                    self.password = keyring.get_password(self.SERVICE_NAME, self.username)
                    while not self.password:
                        self.password = getpass.getpass('Enter your NIMH Data Archives password:')
                    while not self.is_valid_nda_credentials():
                        logger.error("The password that was entered for user '%s' is invalid ...", self.username)
                        logger.error('If your username was previously entered incorrectly, you may update it in your '
                                     'settings.cfg located at \n{}'.format(NDATools.NDA_TOOLS_SETTINGS_CFG_FILE))
                        self.password = getpass.getpass('Enter your NIMH Data Archives password:')
                        while self.password == '':
                            self.password = getpass.getpass('Enter your NIMH Data Archives password:')
                    else:
                        keyring.set_password(self.SERVICE_NAME, self.username, self.password)
                        break
                except RuntimeError as e:
                    if platform.system() == 'Linux':
                        print('If there is no backend set up for keyring, you may try\n'
                              'pip install secretstorage --upgrade keyrings.alt')
                    raise e

        # Only ask for access-key/secret-key when needed (which is only when a user is creating a submission
        # and the files are stored in a s3 bucket)
        if hasattr(self, 'source_bucket') and self.source_bucket:
            self.read_aws_credentials()

    def read_aws_credentials(self):
        if not self.aws_access_key:
            self.aws_access_key = getpass.getpass(
                "Enter your aws_access_key (must have read access to the %s bucket): " % self.source_bucket)

        if not self.aws_secret_key:
            self.aws_secret_key = getpass.getpass('Enter your aws_secret_key:')

    def is_valid_nda_credentials(self):
        try:
            # will raise HTTP error 401 if invalid creds
            Utils.get_request(self.user_api, headers={'content-type': 'application/json'},
                              auth=requests.auth.HTTPBasicAuth(self.username, self.password),
                              error_handler=HttpErrorHandlingStrategy.reraise_status)
            return True
        except HTTPError as e:
            if e.response.status_code == 401:
                if 'locked' in e.response.text:
                    # user account is locked
                    tmp = json.loads(e.response.text)
                    logger.error('\nError: %s', tmp['message'])
                    logger.error('\nPlease contact NDAHelp@mail.nih.gov for help in resolving this error')
                    exit_error()
                else:
                    return False
            else:
                logger.error('\nSystemError while checking credentials for user %s', self.username)
                logger.error('\nPlease contact NDAHelp@mail.nih.gov for help in resolving this error')
                exit_error()



from __future__ import print_function

import json
import os
import pathlib
import shutil
import sys

__version__ = '0.3.0'

from pkg_resources import resource_filename


pypi_version = None
initialization_complete = False
print('Running NDATools Version {}'.format(__version__))


def check_version():
    global pypi_version, initialization_complete
    import requests
    try:
        from packaging.version import parse
    except ImportError:
        from pip._vendor.packaging.version import parse
    # use https://test.pypi.org/pypi/{package}/json on test/release branches, use https://pypi.org on master
    # Print Python 2 support dropping warning
    if sys.version_info < (3, 0):
        print()
        print('''WARNING - Detected Python version 2. Support for Python 2 is being removed from nda-tools. It is recommended to upgrade to the latest version of Python 
          before using the latest features in the downloadcmd''')

    if parse(__version__).is_devrelease:
        return
    url_pattern = 'https://pypi.org/pypi/{package}/json'
    package = 'nda-tools'
    """Return version of package on pypi.python.org using json."""
    req = requests.get(url_pattern.format(package=package))
    version = parse('0')
    if req.status_code == requests.codes.ok:
        j = json.loads(req.text)
        releases = j.get('releases', [])
        for release in releases:
            ver = parse(release)
            if not ver.is_prerelease:
                version = max(version, ver)
    pypi_version = str(version)

    if parse(__version__) < parse(pypi_version):
        print(
            "Your version of nda-tools is out of date. Please upgrade to the latest version ({}) from PyPi or GitHub and "
            "try again. \n\tTo upgrade using pip, run: \r\npip install nda-tools=={}".format(pypi_version,
                                                                                             pypi_version))
        sys.exit(1)


NDA_ORGINIZATION_ROOT_FOLDER = os.path.join(os.path.expanduser('~'), 'NDA')
NDA_TOOLS_ROOT_FOLDER = os.path.join(NDA_ORGINIZATION_ROOT_FOLDER, 'nda-tools')
NDA_TOOLS_VTCMD_FOLDER = os.path.join(NDA_TOOLS_ROOT_FOLDER, 'vtcmd')
NDA_TOOLS_DOWNLOADCMD_FOLDER = os.path.join(
    NDA_TOOLS_ROOT_FOLDER, 'downloadcmd')
NDA_TOOLS_DOWNLOADS_FOLDER = os.path.join(
    NDA_TOOLS_DOWNLOADCMD_FOLDER, 'packages')
NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = os.path.join(
    NDA_TOOLS_DOWNLOADCMD_FOLDER, 'logs')
NDA_TOOLS_VTCMD_LOGS_FOLDER = os.path.join(NDA_TOOLS_VTCMD_FOLDER, 'logs')
NDA_TOOLS_VAL_FOLDER = os.path.join(
    NDA_TOOLS_VTCMD_FOLDER, 'validation_results')
NDA_TOOLS_SUB_PACKAGE_FOLDER = os.path.join(
    NDA_TOOLS_VTCMD_FOLDER, 'submission_package')

NDA_TOOLS_PACKAGE_FILE_METADATA_TEMPLATE = 'package_file_metadata_%s.txt'
NDA_TOOLS_DEFAULT_LOG_FORMAT = '%(asctime)s:%(levelname)s:%(message)s'
NDA_TOOLS_SETTINGS_FOLDER = os.path.join(os.path.expanduser('~'), '.NDATools')
NDA_TOOLS_LOGGING_YML_FILE = os.path.join(NDA_TOOLS_SETTINGS_FOLDER, 'logging.yml')
NDA_TOOLS_SETTINGS_CFG_FILE = os.path.join(NDA_TOOLS_SETTINGS_FOLDER, 'settings.cfg')


def create_nda_folders():
    # init folder structure for program runtime files
    if not os.path.exists(NDA_ORGINIZATION_ROOT_FOLDER):
        os.mkdir(NDA_ORGINIZATION_ROOT_FOLDER)

    if not os.path.exists(NDA_TOOLS_ROOT_FOLDER):
        os.mkdir(NDA_TOOLS_ROOT_FOLDER)

    if not os.path.exists(NDA_TOOLS_VTCMD_FOLDER):
        os.mkdir(NDA_TOOLS_VTCMD_FOLDER)

    if not os.path.exists(NDA_TOOLS_DOWNLOADCMD_FOLDER):
        os.mkdir(NDA_TOOLS_DOWNLOADCMD_FOLDER)

    if not os.path.exists(NDA_TOOLS_DOWNLOADS_FOLDER):
        os.mkdir(NDA_TOOLS_DOWNLOADS_FOLDER)

    if not os.path.exists(NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER):
        os.mkdir(NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER)

    if not os.path.exists(NDA_TOOLS_VTCMD_LOGS_FOLDER):
        os.mkdir(NDA_TOOLS_VTCMD_LOGS_FOLDER)

    if not os.path.exists(NDA_TOOLS_VAL_FOLDER):
        os.mkdir(NDA_TOOLS_VAL_FOLDER)

    if not os.path.exists(NDA_TOOLS_SUB_PACKAGE_FOLDER):
        os.mkdir(NDA_TOOLS_SUB_PACKAGE_FOLDER)

    if not os.path.exists(NDA_TOOLS_SETTINGS_FOLDER):
        os.mkdir(NDA_TOOLS_SETTINGS_FOLDER)

    if not pathlib.Path(NDA_TOOLS_LOGGING_YML_FILE).is_file():
        shutil.copyfile(resource_filename(__name__, 'clientscripts/config/logging.yml'),
                        NDA_TOOLS_LOGGING_YML_FILE)

    if not pathlib.Path(NDA_TOOLS_SETTINGS_CFG_FILE).is_file():
        shutil.copyfile(resource_filename(__name__, 'clientscripts/config/settings.cfg'),
                        NDA_TOOLS_SETTINGS_CFG_FILE)
    # MAC users sometimes see output from python warnings module. Suppress these msgs
    os.environ['PYTHONWARNINGS'] = 'ignore'


def prerun_checks_and_setup():
    check_version()
    create_nda_folders()


def init_and_create_configuration(args, logs_folder , auth_req=True):
    from NDATools.Configuration import ClientConfiguration, LoggingConfiguration
    prerun_checks_and_setup()
    LoggingConfiguration.load_config(logs_folder, args.verbose, args.log_dir)
    config = ClientConfiguration(args)
    config.read_user_credentials(auth_req)
    return config



import sys
from unittest import mock

import keyring
import pytest

from NDATools.Configuration import ClientConfiguration
from NDATools.clientscripts.downloadcmd import configure as download_configure, parse_args as download_parse_args
from NDATools.clientscripts.vtcmd import init_and_create_configuration as validation_configure, \
    parse_args as validation_parse_args


# prevent tests from making any real http requests
@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    """Remove requests.sessions.Session.request for all tests."""
    monkeypatch.delattr("requests.sessions.Session.request")

def mock_get_password(*args, **kwargs):
    return 'fake-pass'
def mock_is_valid_credentials(*args, **kwargs):
    return True
@pytest.fixture
def download_config_factory(monkeypatch):
    def _make_config(test_args):
        with monkeypatch.context() as m:
            test_args.insert(0, 'downloadcmd')
            m.setattr(sys, 'argv', test_args)
            m.setattr(keyring, 'get_password', mock_get_password)
            m.setattr(ClientConfiguration, 'is_valid_nda_credentials', mock_is_valid_credentials)
            args = download_parse_args()
            config = download_configure(args)
            return args, config
    return _make_config


@pytest.fixture
def validation_config_factory():
    def _make_val_config(test_args):
        with mock.patch.object(sys, 'argv', test_args):
            test_args.insert(0, 'vtcmd')
            args = validation_parse_args()
            config = validation_configure(args)
        return args, config

    return _make_val_config


@pytest.fixture
def load_from_file(shared_datadir):
    def _load_from_file(file):
        content = (shared_datadir / file).read_text()
        return content

    return _load_from_file


import json
import uuid
from unittest.mock import MagicMock

import pytest

from NDATools import Submission, Utils
from NDATools.Submission import Status


class TestSubmission():

    @pytest.fixture
    def api_endpoint(self):
        def fix_factory(*args, **kwargs):
            api_response = args[0]
            def __api_result(*args, **kwargs):
                return api_response
            return __api_result
        return fix_factory

    @pytest.fixture
    def config(self, shared_datadir, validation_config_factory):
        file_path = (shared_datadir / 'validation/file.csv')
        test_args = [str(file_path)]
        _, config = validation_config_factory(test_args)
        config.JSON = True
        return config

    @pytest.fixture
    def submission_with_files(self, load_from_file):
        files_resource = json.loads(load_from_file('submission/api_response/get_files_response.json'))
        submission_resource =  json.loads(load_from_file('submission/api_response/create_submission_response.json'))
        creds_resource =  json.loads(load_from_file('submission/api_response/batch_get_temp_creds.json'))
        return (submission_resource, files_resource, creds_resource)

    @pytest.fixture
    def batch_update_status(self, load_from_file):
        return json.loads(load_from_file('submission/api_response/batch_update_status_response.json'))

    def test_resume_submission(self, monkeypatch, config, submission_with_files, api_endpoint, tmpdir, batch_update_status):
        submission_resource, files_resource, creds_resource = submission_with_files
        file_key = Utils.sanitize_file_path(list(filter(lambda x: x['file_type']=='Submission Associated File', files_resource))[0]['file_user_path'])
        fake_file_size = 10
        fake_abs_path = str(tmpdir/file_key)
        with monkeypatch.context() as m:
            mock_status = MagicMock()
            mock_get_creds = MagicMock(return_value =creds_resource['credentials'])
            mock_batch_update = MagicMock(return_value =batch_update_status)
            mock_abort_previous_uploads = MagicMock()
            mock_upload_files = MagicMock()
            m.setattr(Submission.Submission, 'check_status', mock_status)
            m.setattr(Submission.Submission, 'get_multipart_credentials', mock_get_creds)
            m.setattr(Submission.Submission, 'abort_previous_upload_attempts', mock_abort_previous_uploads)
            m.setattr(Submission.Submission, 'batch_update_status', mock_batch_update['errors'])
            m.setattr(Submission.Submission, 'upload_associated_files', mock_upload_files)
            m.setattr(Submission, 'post_request', api_endpoint(submission_resource))
            m.setattr(Submission, 'get_request', api_endpoint(files_resource))
            def mock_complicated_search_file_logic(*args, **kwargs):
                submission.full_file_path = { file_key: (fake_abs_path, fake_file_size)}
                args[1].clear()
                return
            mock_search_file_sys = MagicMock(wraps=mock_complicated_search_file_logic)
            m.setattr(Submission, 'parse_local_files',mock_search_file_sys)

            submission = Submission.Submission(submission_id=submission_resource['submission_id'],
                                               full_file_path={},
                                               thread_num=1,
                                               batch_size=20,
                                               allow_exit=True,
                                               config=config)
            submission.status = Status.UPLOADING
            submission.directory_list = [str(tmpdir)]
            submission.resume_submission()

            assert mock_status.call_count == 2
            assert mock_upload_files.call_count == 1
            assert mock_get_creds.call_count == 1
            assert mock_abort_previous_uploads.call_count ==1 
            assert submission.get_files()==files_resource

    def test_create_submission_success(self, monkeypatch, config, submission_with_files, api_endpoint):
        submission_resource, files_resource, _ = submission_with_files

        with monkeypatch.context() as m:
            m.setattr(Submission, 'post_request', api_endpoint(submission_resource))
            m.setattr(Submission, 'get_request', api_endpoint(files_resource))
            package_id = uuid.uuid4()

            submission = Submission.Submission(package_id=str(package_id),
                                                         full_file_path={},
                                                         thread_num=1,
                                                         batch_size=20,
                                                         allow_exit=True,
                                                         config=config)
            submission.submit()

            assert submission.submission_id == submission_resource['submission_id']
            assert submission.status == submission_resource['submission_status']
            assert submission.get_files()==files_resource

    def test_replace_submission(self, monkeypatch, config, submission_with_files, api_endpoint):

        submission_resource, files_resource, _ = submission_with_files

        with monkeypatch.context() as m:
            m.setattr(Submission, 'put_request', api_endpoint(submission_resource))
            m.setattr(Submission, 'get_request', api_endpoint(files_resource))
            package_id = uuid.uuid4()

            submission = Submission.Submission(submission_id=submission_resource['submission_id'],
                                               package_id=str(package_id),
                                               full_file_path={},
                                               thread_num=1,
                                               batch_size=20,
                                               allow_exit=True,
                                               config=config)
            submission.replace_submission()

            assert submission.submission_id == submission_resource['submission_id']
            assert submission.status == submission_resource['submission_status']
            assert submission.get_files()==files_resource




import logging
import os

from NDATools.Utils import parse_local_files, sanitize_file_path, check_read_permissions, sanitize_windows_download_filename
from unittest import TestCase
from mock import patch
import mock

logging.basicConfig(level=logging.DEBUG, format="%(asctime)s:%(levelname)s:%(message)s")
logger = logging.getLogger(__name__)


class TestUtils(TestCase):

    def test_sanitize_file_path(self):
        windows_path = r'C:\\Users\\test\\path\\file.csv'
        linux_path = '/home/test/path/file.csv'
        self.assertEqual(sanitize_file_path(windows_path), 'Users/test/path/file.csv')
        self.assertEqual(sanitize_file_path(linux_path), 'home/test/path/file.csv')

    def test_sanitize_windows_download_filename(self):
        path_containing_prohibited_windows_char = r'C:\\Users\\test\\path><\\package|15?*15\\file:12:22.csv'
        self.assertEqual(r'C%3A\\Users\\test\\path%3E%3C\\package%7C15%3F%2A15\\file%3A12%3A22.csv',
                         sanitize_windows_download_filename(path_containing_prohibited_windows_char))

    @patch('os.path.getsize')
    @patch('os.path.isfile')
    @patch('NDATools.Utils.check_read_permissions')
    def test_parse_local_files(self, mock_file_size, mock_isfile, mock_check_read_permissions):

        mock_file_size.return_value = 1
        mock_isfile.return_value = True
        mock_check_read_permissions.return_value = True

        # Test relative paths
        windows_relative_path = 'windows_file.csv'
        linux_relative_path = 'linux_file.csv'
        mac_relative_path = 'mac_file.csv'

        file_list = [windows_relative_path, linux_relative_path, mac_relative_path]

        windows_directory = r'C:\Users\test\path'
        linux_directory = '/home/user/path'
        mac_directory = '/Users/user/test'

        directory_list = [windows_directory, linux_directory, mac_directory]

        file_size_full_path_dict = {}
        invalid_file_array = []

        parse_local_files(directory_list, file_list, file_size_full_path_dict, invalid_file_array, True)

        logger.debug(file_list)
        logger.debug(invalid_file_array)
        logger.debug(file_size_full_path_dict)

        self.assertEqual(len(invalid_file_array), 0)
        self.assertEqual(len(file_list), 0)
        self.assertEqual(len(file_size_full_path_dict), 3)

        # Test full paths
        windows_full_path = r'C:\Users\test\path\windows_file_full.csv'
        linux_full_path = '/home/user/path/linux_file_full.csv'
        mac_full_path = '/Users/user/test/mac_file_full.csv'

        file_list = [windows_full_path, linux_full_path, mac_full_path]

        parse_local_files(directory_list, file_list, file_size_full_path_dict, invalid_file_array, False)

        logger.debug(file_list)
        logger.debug(invalid_file_array)
        logger.debug(file_size_full_path_dict)

        self.assertEqual(len(invalid_file_array), 0)
        self.assertEqual(len(file_list), 0)
        self.assertEqual(len(file_size_full_path_dict), 6)

    @patch('builtins.open', mock.mock_open(read_data=''))
    def test_check_read_permissions(self):
        test_file = os.path.join(os.path.expanduser('~'), 'NDATools\\clientscripts\\config\\settings.cfg')
        self.assertTrue(check_read_permissions(test_file))

    @patch('builtins.open')
    def test_check_read_permissions_denied(self, mock_file):
        mock_file.side_effect = IOError()
        test_file = os.path.join(os.path.expanduser('~'), 'NDATools\\clientscripts\\config\\settings.cfg')
        self.assertFalse(check_read_permissions(test_file))



import datetime
import itertools
import json
import math
import os
import shutil
import sys
from unittest.mock import ANY, MagicMock

import pytest
from requests import HTTPError

import NDATools.Utils
import NDATools.clientscripts.downloadcmd
from NDATools.Download import Download

FAKE_FILE_BYTES = [0x10, 0x10]
MISSING_FILE = 's3://NDAR_Central_1/submission_43568/not-found.png'


class TestDownload:

    @pytest.fixture(autouse=True)
    def class_setup(self, load_from_file):
        self.load_from_file = load_from_file


    def test_s3links_argument_file_not_found(self,
                                        download_config_factory,
                                        shared_datadir,
                                        capsys,
                                        monkeypatch,
                                        tmp_path,
                                        tmpdir):
        test_package_id = '1189934'
        test_text_structure_file = 'api_responses/s3/ds_test/test-not-found.csv'
        test_args = ['-dp', test_package_id, '-t', os.path.join(shared_datadir, test_text_structure_file)]

        config, args = download_config_factory(test_args)

        def mock_initialize_verification_files(*args, **kwargs):
            self = args[0]
            meta = tmpdir / "meta"
            meta.mkdir()
            self.package_metadata_directory = str(meta)
            logs = tmpdir / "logs"
            logs.mkdir()
            NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = str(logs)
            return str(tmpdir/ "test_dd.csv")
        def mock_get_package_info(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-package-info-response.json'))

        def mock_generate_download_batch_file_ids(*args, **kwargs):
            return []

        def mock_get_completed_files_in_download(*args, **kwargs):
            return []

        download_mock = MagicMock(return_value = {'actual_file_size':10, 'download_complete_time': datetime.datetime.now()})

        def get_presigned_urls_mock(*args, **kwargs):
            return {f:'s3://fake-presigned-url' for f in args[1]}


        def mock_post_request (*args, **kwargs):
            response = MagicMock()
            payload = args[1]
            if  MISSING_FILE in payload:
                response_text = '''The following files are invalid\r\n{}'''.format(MISSING_FILE)
                response.text = response_text
                response.status_code = 404
                e = HTTPError(response=response)

                def rethrow(*args, **kwargs):
                    raise e

                response.ok.return_value = False
                response.raise_for_status.side_effect = rethrow
                return response
            else:
                response_text = self.load_from_file('api_responses/package/ds_test/get_files_by_s3.json')
                response.json.return_value = json.loads(response_text)
                response.text = response_text
                return response

        with monkeypatch.context() as m:
            m.setattr(Download, 'initialize_verification_files', mock_initialize_verification_files)
            m.setattr(Download, 'get_package_info', mock_get_package_info)
            m.setattr(Download, 'get_completed_files_in_download', mock_get_completed_files_in_download)
            m.setattr(NDATools.Download, 'post_request', mock_post_request)
            m.setattr(Download, 'download_from_s3link', download_mock)
            m.setattr(Download, 'get_presigned_urls', get_presigned_urls_mock)

            Download(args, config).start()

            captured = capsys.readouterr()
            assert 'WARNING: The following associated files were not found' in captured.out
            assert 'Beginning download of 3 files' in captured.out
            assert download_mock.call_count == 3  # number of files in the DS file


    def test_ds_argument_structure_not_found(self,
                                             download_config_factory,
                                             capsys, monkeypatch, shared_datadir, tmp_path, tmpdir):

        test_package_id = '1189934'
        test_data_structure_file = 'image03'
        test_args = ['-dp', test_package_id, '-ds', test_data_structure_file]

        config, args = download_config_factory(test_args)

        def mock_initialize_verification_files(*args, **kwargs):
            self = args[0]
            meta = tmpdir / "meta"
            meta.mkdir()
            self.package_metadata_directory = str(meta)
            logs = tmpdir / "logs"
            logs.mkdir()
            NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = str(logs)
            return str(tmpdir/ "test_dd.csv")
        def mock_get_package_info(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-package-info-response.json'))
        def mock_generate_download_batch_file_ids(*args, **kwargs):
            return []

        def mock_get_completed_files_in_download(*args, **kwargs):
            return []

        def mock_os_exit(*args, **kwargs):
            sys.exit()
        def mock_get_files_from_datastructure(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/package/ds_test/no-files.json'))['results']

        with monkeypatch.context() as m:
            m.setattr(Download, 'initialize_verification_files', mock_initialize_verification_files)
            m.setattr(Download, 'get_package_info', mock_get_package_info)
            m.setattr(Download, 'get_completed_files_in_download', mock_get_completed_files_in_download)
            m.setattr(Download, 'get_files_from_datastructure', mock_get_files_from_datastructure)

            m.setattr(Download, 'generate_download_batch_file_ids', mock_generate_download_batch_file_ids)
            m.setattr(os, '_exit', mock_os_exit)


            with pytest.raises(SystemExit):
                Download(args, config).start()

            captured = capsys.readouterr()
            # Program must print out that the structure was not found
            assert "{} data structure is not included in the package".format(test_data_structure_file) in captured.out


    def test_ds_argument_success(self,
                                 download_config_factory,
                                 shared_datadir, tmp_path, tmpdir, monkeypatch):

        test_package_id = '1189934'
        test_data_structure_file = 'fmriresults01'
        test_args = ['-dp', test_package_id, '-ds', test_data_structure_file, '-d', str(tmpdir)]

        config, args = download_config_factory(test_args)
        def get_presigned_urls_mock(*args, **kwargs):
            return {f:'s3://fake-presigned-url' for f in args[1]}

        def mock_initialize_verification_files(*args, **kwargs):
            self = args[0]
            meta = tmpdir / "meta"
            meta.mkdir()
            self.package_metadata_directory = str(meta)
            logs = tmpdir / "logs"
            logs.mkdir()
            NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = str(logs)
            return str(tmpdir / "test_dd.csv")

        def mock_get_package_info(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-package-info-response.json'))

        def mock_get_files_from_datastructure(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/package/ds_test/all_mri_files.json'))['results']

        def mock_get_completed_files_in_download(*args, **kwargs):
            return []

        download_mock = MagicMock(return_value = {'actual_file_size':10, 'download_complete_time': datetime.datetime.now()})
        with monkeypatch.context() as m:
            m.setattr(Download, 'initialize_verification_files', mock_initialize_verification_files)
            m.setattr(Download, 'get_package_info', mock_get_package_info)
            m.setattr(Download, 'get_completed_files_in_download', mock_get_completed_files_in_download)
            m.setattr(Download, 'use_data_structure', mock_get_files_from_datastructure)
            m.setattr(Download, 'download_from_s3link', download_mock)
            m.setattr(Download, 'get_presigned_urls', get_presigned_urls_mock)

            Download(args, config).start()
            # write should be called for each fake byte in the file (see above mock_api_response call)

            assert download_mock.call_count == 3  # number of files in the DS file


    def test_resume_download_using_download_history_report(self,
                                 download_config_factory,
                                 shared_datadir, tmp_path, tmpdir, monkeypatch):

        test_package_id = '1189934'
        test_data_structure_file = 'fmriresults01'
        test_args = ['-dp', test_package_id, '-ds', test_data_structure_file, '-d', str(tmpdir)]

        config, args = download_config_factory(test_args)
        def get_presigned_urls_mock(*args, **kwargs):
            return {f:'s3://fake-presigned-url' for f in args[1]}

        def mock_initialize_verification_files(*args, **kwargs):
            self = args[0]
            meta = tmpdir / "meta"
            meta.mkdir()
            self.package_metadata_directory = str(meta)
            logs = tmpdir / "logs"
            logs.mkdir()
            NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = str(logs)
            self.download_job_uuid = '9479421578f74b8997f62cd962a1cb08'

            tmp1 = meta / '.download-progress'
            tmp1.mkdir()

            tmp2 = tmp1 / self.download_job_uuid
            tmp2.mkdir()

            shutil.copy(os.path.join(shared_datadir, 'api_responses/package/ds_test/download-progress-report.csv'), str(tmp2))

            return str(tmpdir / "test_dd.csv")

        def mock_get_package_info(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-package-info-response.json'))

        def mock_get_files_from_datastructure(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/package/ds_test/all_mri_files.json'))['results']

        download_mock = MagicMock(return_value = {'actual_file_size':10, 'download_complete_time': datetime.datetime.now()})
        with monkeypatch.context() as m:
            m.setattr(Download, 'initialize_verification_files', mock_initialize_verification_files)
            m.setattr(Download, 'get_package_info', mock_get_package_info)

            m.setattr(Download, 'use_data_structure', mock_get_files_from_datastructure)
            m.setattr(Download, 'download_from_s3link', download_mock)
            m.setattr(Download, 'get_presigned_urls', get_presigned_urls_mock)

            Download(args, config).start()

            assert download_mock.call_count == 1  # number of files in the DS file
            download_mock.assert_called_with(3132764123, 's3://fake-presigned-url', failed_s3_links_file=ANY)


    def test_s3links_argument_success(self,
                                      download_config_factory,
                                      shared_datadir,
                                      capsys,
                                      monkeypatch,
                                      tmp_path,
                                      tmpdir):
        test_package_id = '1189934'
        test_text_structure_file = 'api_responses/s3/ds_test/test-found.csv'
        test_args = ['-dp', test_package_id, '-t', os.path.join(shared_datadir, test_text_structure_file)]

        config,args = download_config_factory(test_args)

        def mock_initialize_verification_files(*args, **kwargs):
            self = args[0]
            meta = tmpdir / "meta"
            meta.mkdir()
            self.package_metadata_directory = str(meta)
            logs = tmpdir / "logs"
            logs.mkdir()
            NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = str(logs)
            return str(tmpdir / "test_dd.csv")

        def mock_get_package_info(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-package-info-response.json'))

        def mock_generate_download_batch_file_ids(*args, **kwargs):
            return []

        def mock_get_completed_files_in_download(*args, **kwargs):
            return []
        def mock_use_s3_link_file(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/package/ds_test/get_files_by_s3.json'))

        with monkeypatch.context() as m:
            m.setattr(Download, 'initialize_verification_files', mock_initialize_verification_files)
            m.setattr(Download, 'get_package_info', mock_get_package_info)
            m.setattr(Download, 'get_completed_files_in_download', mock_get_completed_files_in_download)


            m.setattr(Download, 'use_s3_links_file', mock_use_s3_link_file)
            m.setattr(Download, 'generate_download_batch_file_ids', mock_generate_download_batch_file_ids)

            # m.setattr(os, 'path.sep', self.mock_post_api_response)
            Download(args, config).start()

            captured = capsys.readouterr()
            assert 'WARNING: The following associated files were not found' not in captured.out
            assert 'Beginning download of 3 files' in captured.out

    def test_generate_s3_files_batch(self,
                                      download_config_factory,
                                      shared_datadir,
                                      capsys,
                                      monkeypatch,
                                      tmp_path,
                                      tmpdir):


        def mock_initialize_verification_files(*args, **kwargs):
            self = args[0]
            self.default_download_batch_size=TEST_BATCH_SIZE
            meta = tmpdir / "meta"
            if not os.path.exists(meta):
                meta.mkdir()
            self.package_metadata_directory = str(meta)
            self.download_job_uuid = '123e4567-e89b-12d3-a456-426614174000'
            logs = tmpdir / "logs"
            if not os.path.exists(logs):
                logs.mkdir()
            NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER = str(logs)
            return str(tmpdir / "test_dd.csv")

        def get_presigned_urls_mock(*args, **kwargs):
            return {f:'s3://fake-presigned-url' for f in args[0]}

        def mock_get_package_info(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-package-info-response.json'))

        def mock_get_completed_files_in_download(*args, **kwargs):
            def _mock_fn(*args2, **kwargs2):
                return [{'package_file_id':int(r), 'actual_file_size': 10, 'download_complete_time': datetime.datetime.now()} for r in kwargs['completed_files']]
            return _mock_fn

        def get_package_files_by_page_method_mock(*args, **kwargs):
            page=args[0]
            batch_size = args[1]
            res = json.loads(self.load_from_file('api_responses/s3/ds_test/get-files-from-datastructure-response.json'))[
                'results']
            from_i = ((page-1) * batch_size)
            to_i = from_i + batch_size
            return res[from_i: to_i]

        def mock_use_datastructure(*args, **kwargs):
            return json.loads(self.load_from_file('api_responses/s3/ds_test/get-files-from-datastructure-response.json'))[
                'results']

        all_files = json.loads(self.load_from_file('api_responses/s3/ds_test/get-files-from-datastructure-response.json'))['results']
        all_file_ids = set(map(lambda x: int(x['package_file_id']), all_files))

        def test(config,args, batch_size, completed_files, get_package_files_by_page_args_list):
            completed_file_ids = set(map(lambda x: int(x['package_file_id']), completed_files))
            with monkeypatch.context() as m:
                m.setattr(Download, 'initialize_verification_files', mock_initialize_verification_files)
                m.setattr(Download, 'get_package_info', mock_get_package_info)
                m.setattr(Download, 'get_completed_files_in_download', mock_get_completed_files_in_download(completed_files=completed_file_ids))

                #mock_presigned_url_method = MagicMock()
                mock_download_method = MagicMock(return_value={'actual_file_size':10, 'download_complete_time': datetime.datetime.now()})
                m.setattr(Download, 'download_from_s3link', mock_download_method)

                d = Download(args, config)
                wrap_download_batch_file_ids = MagicMock(wraps=d.generate_download_batch_file_ids)
                m.setattr(Download, 'generate_download_batch_file_ids', wrap_download_batch_file_ids)
                m.setattr(d, 'default_download_batch_size', batch_size)
                get_presigned_wrap = MagicMock(wraps=get_presigned_urls_mock)
                m.setattr(Download, 'get_presigned_urls', get_presigned_wrap)
                m.setattr(Download, 'use_data_structure', mock_use_datastructure)

                mock_get_package_files_by_page_method = MagicMock(wraps=get_package_files_by_page_method_mock)
                m.setattr(Download, 'get_package_files_by_page', mock_get_package_files_by_page_method)

                d.start()

                assert wrap_download_batch_file_ids.call_count == 1 # its a generator method so it will always be called once, although it may yield many values
                assert list(map(lambda x: x[0], mock_get_package_files_by_page_method.call_args_list)) == get_package_files_by_page_args_list
                assert mock_download_method.call_count == len(all_file_ids) - len(completed_file_ids) # call download method for each file that hasnt been downloaded
                assert set(map(lambda x: x[0][0], mock_download_method.call_args_list)) == {f for f in all_file_ids if f not in completed_file_ids}
                # we started batching calls to this endpoint, so it would be the
                if d.download_mode=='package':
                    assert get_presigned_wrap.call_count == len(get_package_files_by_page_args_list) - 1 == math.ceil((len(all_file_ids) - len(completed_file_ids)) / TEST_BATCH_SIZE)
                    assert set(d.local_file_names.keys()) == all_file_ids - completed_file_ids
                else:
                    assert get_presigned_wrap.call_count == math.ceil((len(all_file_ids) - len(completed_file_ids)) / TEST_BATCH_SIZE)
                    assert set(d.local_file_names.keys()) == all_file_ids #
                assert set(itertools.chain(*map(lambda x: x[0][0], get_presigned_wrap.call_args_list))) == {f for f in all_file_ids if f not in completed_file_ids}


        test_package_id = '1189934'
        test_args = ['-dp', test_package_id, '-u', 'ndar_administrator']
        config,args = download_config_factory(test_args)
        TEST_BATCH_SIZE = 3

        # test package download
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=[], get_package_files_by_page_args_list=[(1, TEST_BATCH_SIZE), (2, TEST_BATCH_SIZE), (3, TEST_BATCH_SIZE)])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:1], get_package_files_by_page_args_list=[(1, TEST_BATCH_SIZE), (2, TEST_BATCH_SIZE), (3, TEST_BATCH_SIZE)])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:2], get_package_files_by_page_args_list=[(1, TEST_BATCH_SIZE), (2, TEST_BATCH_SIZE), (3, TEST_BATCH_SIZE)])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:3], get_package_files_by_page_args_list=[(2, TEST_BATCH_SIZE), (3, TEST_BATCH_SIZE)])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:4], get_package_files_by_page_args_list=[(2, TEST_BATCH_SIZE), (3, TEST_BATCH_SIZE)])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:5], get_package_files_by_page_args_list=[(2, TEST_BATCH_SIZE), (3, TEST_BATCH_SIZE)])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:], get_package_files_by_page_args_list=[(3, TEST_BATCH_SIZE)])

        # test download ds
        test_package_id = '1189934'
        test_args = ['-dp', test_package_id, '-ds', 'fmriresults01', '-u', 'ndar_administrator']
        config,args = download_config_factory(test_args)
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=[], get_package_files_by_page_args_list=[])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:1], get_package_files_by_page_args_list=[])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:2], get_package_files_by_page_args_list=[])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:3], get_package_files_by_page_args_list=[])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:4], get_package_files_by_page_args_list=[])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:5], get_package_files_by_page_args_list=[])
        test(config,args, batch_size=TEST_BATCH_SIZE, completed_files=all_files[:], get_package_files_by_page_args_list=[])



from __future__ import absolute_import, with_statement

import sys

import requests.packages.urllib3.util

if sys.version_info[0] < 3:
    input = raw_input
import requests.packages.urllib3.util
from NDATools.Configuration import *
from NDATools.Utils import *

logger = logging.getLogger(__name__)


class SubmissionPackage:
    def __init__(self, uuid, associated_files, config, pending_changes=None, original_uuids=None):
        self.config = config
        self.api = self.config.submission_package_api
        self.validationtool_api = self.config.validationtool_api
        self.uuid = uuid
        self.associated_files = associated_files
        self.username = self.config.username
        self.password = self.config.password
        self.dataset_name = self.config.title
        self.dataset_description = self.config.description
        self.package_info = {}
        self.download_links = []
        self.package_id = None
        self.package_folder = None
        self.auth = requests.auth.HTTPBasicAuth(self.config.username, self.config.password)
        self.alt_endpoints = None
        self.collection_id = self.check_and_get_collection_id()
        self.validation_results = []
        self.pending_changes = pending_changes
        self.original_validation_uuids = original_uuids
        # these fields will be set after the package is created
        self.expiration_date = None
        self.create_date = None
        self.submission_package_uuid = None

    def get_collections(self):
        collections = get_request("/".join([self.validationtool_api, "user/collection"]), auth=self.auth,
                                  headers={'Accept': 'application/json'})
        return {int(c['id']): c['title'] for c in collections}

    def has_permissions_to_submit_to_collection(self, collection_id, user_collections):
        c_id = int(collection_id)
        if c_id in user_collections:
            return True
        # TODO refactor collection-api to remove the need to make these separate api calls
        if self.alt_endpoints is not None:
            return c_id in self.alt_endpoints
        else:
            endpoints = get_request("/".join([self.validationtool_api, "user/customEndpoints"]), auth=self.auth,
                                    headers={'Accept': 'application/json'})
            # TODO cache api response
            all_collections = get_request("/".join([self.config.collection_api]), auth=self.auth,
                                          headers={'Accept': 'application/json'})
            tmp_endpoints = {e['title'] for e in endpoints}
            self.alt_endpoints = {int(c['id']): c['title'] for c in all_collections if
                                  c['altEndpoint'] in tmp_endpoints}
            return c_id in self.alt_endpoints

    def check_and_get_collection_id(self):
        user_collections = self.get_collections()
        if not user_collections:
            message = 'The user {} does not have permission to submit to any collections.'.format(self.config.username)
            exit_error(message=message)
        if self.config.collection_id:
            if not self.has_permissions_to_submit_to_collection(self.config.collection_id, user_collections):
                message = 'The user {} does not have permission to submit to collection {}.'.format(
                    self.config.username, self.config.collection_id)
                exit_error(message=message)
            else:
                return self.config.collection_id
        else:
            return self.prompt_for_collection_id(user_collections)

    def prompt_for_collection_id(self, user_collections):
        while True:
            try:
                user_input = int(input('\nEnter collection ID:').strip())
                if not self.has_permissions_to_submit_to_collection(user_input, user_collections):
                    logger.error(f'You do not have access to submit to the collection: {user_input} ')
                    logger.info(f'Please choose from one of the following collections: ')
                    for collection_id in sorted(user_collections.keys()):
                        logger.info('{}: {}'.format(collection_id, user_collections[collection_id]))
                else:
                    return user_input
            except ValueError:
                logger.error('Error: Input must be a valid integer')

    def build_package(self):
        def raise_error(value):
            raise Exception("Missing {}. Please try again.".format(value))

        if not self.dataset_name:
            raise_error('dataset name')

        if not self.dataset_description:
            raise_error('dataset description')

        if self.collection_id is None:
            raise_error('collection ID')

        self.package_info = {
            "package_info": {
                "dataset_description": self.dataset_description,
                "dataset_name": self.dataset_name,
                "collection_id": self.collection_id
            },
            "validation_results":
                self.uuid
        }
        if self.config.replace_submission:
            self.package_info['package_info']['replacement_submission'] = self.config.replace_submission
            self.print_replacement_summary()
            if not self.config.force:
                user_input = evaluate_yes_no_input("Are you sure you want to continue?", 'n')
                if user_input.lower() == 'n':
                    exit_error(message='Exiting...')

        json_data = json.dumps(self.package_info)
        response = post_request(self.api, json_data, auth=self.auth)
        if response:
            try:
                self.package_id = response['submission_package_uuid']
                for r in response['validation_results']:
                    self.validation_results.append(r['id'])
                self.submission_package_uuid = str(response['submission_package_uuid'])
                self.create_date = str(response['created_date'])
                self.expiration_date = str(response['expiration_date'])
            except KeyError:
                message = 'There was an error creating your package.'
                if response['status'] == Status.ERROR:
                    message = response['errors'][0]['message']
                exit_error(message=message)

            while response['package_info']['status'] == Status.PROCESSING:
                time.sleep(1.1)
                response = get_request("/".join([self.api, self.package_id]), auth=self.auth)
            if response['package_info']['status'] == Status.COMPLETE:
                for f in [f for f in response['files']
                          if f['type'] in ('Submission Memento', 'Submission Data Package')]:
                    for key, value in f['_links'].items():
                        for k, v in value.items():
                            self.download_links.append((v, "/".join(f['path'].split('/')[4:])))
            else:
                message = 'There was an error in building your package.'
                if response['package_info']['status'] == Status.SYSERROR:
                    message = response['errors']['system'][0]['message']
                elif 'has changed since validation' in response['errors']:
                    message = response['errors']
                exit_error(message=message)
        else:
            message = 'There was an error with your package request.'
            exit_error(message=message)

    def print_replacement_summary(self):
        logger.info('Below is a summary of what your submission will look like with the validation files provided:')
        logger.info('')
        logger.info('Short-Name, Number of Rows')
        for change in self.pending_changes:
            logger.info('{},{}'.format(change['shortName'], change['rows']))
        logger.info('')
        logger.info('')


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'
    COMPLETE = 'complete'
    ERROR = 'error'
    PROCESSING = 'processing'



import json
from unittest.mock import MagicMock

import pytest

from NDATools import Validation


class TestValidation:

    def _provide_get_responses(self):
        for x in ['validation/api_response/initiate_validation.json',
                  'validation/api_response/initiate_validation.json',
                  'validation/api_response/initiate_validation.json',
                  'validation/api_response/initiate_validation.json',
                  'validation/api_response/initiate_validation.json',
                  'validation/api_response/finish_validation.json',
                  'validation/api_response/finish_validation.json']:
            yield self._load_from_file(x)

    def mocked_api_request_successful(self, *args, **kwargs):
        session = MagicMock()
        verb = args[1]
        endpoint = args[2]
        text = None

        if verb == 'GET':
            text = next(self._get_responses)

        json_object = json.loads(text)

        return json_object


    @pytest.fixture(autouse=True)
    def setup(self, load_from_file):
        self._get_responses = self._provide_get_responses()
        self._load_from_file = load_from_file

    def test_validation_with_scope(self, monkeypatch,
                                   validation_config_factory,
                                   shared_datadir,
                                   tmp_path):

        file_path = (shared_datadir / 'validation/file.csv')

        test_args = ['--scope', '123123', str(file_path)]

        args, config = validation_config_factory(test_args)
        config.JSON = True

        def mocked_post_request(*args, **kwargs):
            return json.loads(self._load_from_file('validation/api_response/initiate_validation.json'))

        def mocked_get_request(*args, **kwargs):
            return json.loads(next(self._get_responses))

        with monkeypatch.context() as m:
            m.setattr(Validation, 'post_request', mocked_post_request)
            m.setattr(Validation, 'get_request', mocked_get_request)
            validation = Validation.Validation(args.files, config=config, hide_progress=config.hideProgress, thread_num=1,
                                    allow_exit=True)
            validation.validate()
            result = validation.responses[0][0]

            assert validation.responses[0][1] == str(file_path)
            assert result['id'] == 'test'
            assert result['status'] == 'Complete'
            assert result['expiration_date'] == '07/14/2021'
            assert result['errors'] == {}



from __future__ import absolute_import, with_statement

import copy
import csv
import gzip
import multiprocessing
import os.path
import pathlib
import platform
import shutil
import tempfile
import uuid
from queue import Queue
from shutil import copyfile
from threading import Thread

import pandas as pd
from boto3.s3.transfer import TransferConfig
from requests import HTTPError

import NDATools
from NDATools import Utils
from NDATools.AltEndpointSSLAdapter import AltEndpointSSLAdapter
from NDATools.Utils import *

logger = logging.getLogger(__name__)



class ThreadPool:
    """ Pool of threads consuming tasks from a queue """

    class Worker(Thread):
        """ Thread executing tasks from a given tasks queue """

        def __init__(self, tasks):
            Thread.__init__(self)
            self.tasks = tasks
            self.daemon = True
            self.start()

        def run(self):
            while True:
                func, args = self.tasks.get()
                try:
                    func(*args)
                except Exception as e:
                    # An exception happened in this thread
                    logger.info(str(e))
                    logger.info(get_traceback())
                finally:
                    # Mark this task as done, whether an exception happened or not
                    self.tasks.task_done()

    def __init__(self, num_threads, queue_size=None):
        queue_size = queue_size or num_threads * 100
        self.tasks = Queue(queue_size)
        for _ in range(num_threads):
            ThreadPool.Worker(self.tasks)

    def map(self, func, args_list):
        """ Add a list of tasks to the queue """
        for args in args_list:
            self.tasks.put((func, args))

    def wait_completion(self):
        """ Wait for completion of all the tasks in the queue """
        self.tasks.join()


class DownloadRequest():

    def __init__(self,  package_file, presigned_url, package_id, download_dir):
        operating_system = platform.system()
        if operating_system == 'Windows':
            download_dir = Utils.sanitize_windows_download_filename(download_dir)
        self.presigned_url = presigned_url
        self.package_file_id = str(package_file['package_file_id'])
        self.package_file_relative_path = package_file['download_alias']
        self.completed_download_abs_path = os.path.normpath(Utils.convert_to_abs_path(os.path.join(download_dir,
                         Utils.sanitize_windows_download_filename(self.package_file_relative_path)
                         if operating_system == 'Windows' else self.package_file_relative_path)
        ))
        self.package_download_directory = Utils.convert_to_abs_path(os.path.join(NDATools.NDA_TOOLS_DOWNLOADS_FOLDER, str(package_id)))
        self.nda_s3_url = None
        self.exists = False
        self.expected_file_size = package_file['file_size']
        self.actual_file_size = 0
        self.e_tag = None,
        self.download_complete_time = None
        self.partial_download_abs_path = self.completed_download_abs_path + '.partial'


class Download(Protocol):

    def __init__(self, download_config, args):

        # Instance variables from config
        self.config = download_config
        self.package_url = self.config.package_api
        self.package_creation_url = self.config.package_creation_api
        self.datadictionary_url = self.config.datadictionary_api
        self.username = download_config.username
        self.password = download_config.password
        self.auth = requests.auth.HTTPBasicAuth(self.config.username, self.config.password)

        # Instance Variables from 'args'
        if args.directory:
            download_directory = args.directory[0]
        else:
            download_directory = os.path.join(NDATools.NDA_TOOLS_DOWNLOADS_FOLDER, str(args.package))
        self.package_metadata_directory = os.path.join(NDATools.NDA_TOOLS_DOWNLOADS_FOLDER,
                                                       str(args.package))
        self.download_directory = Utils.convert_to_abs_path(download_directory)
        self.s3_links_file = args.txt
        self.inline_s3_links = args.paths
        self.package_id = args.package
        self.data_structure = args.datastructure
        self.thread_num = args.workerThreads if args.workerThreads else max([1, multiprocessing.cpu_count() - 1])
        self.regex_file_filter = args.file_regex
        if self.s3_links_file:
            self.download_mode = 'text'
        elif self.data_structure:
            self.download_mode = 'datastructure'
        elif self.inline_s3_links:
            self.download_mode = 'paths'
        else:
            self.download_mode = 'package'
        self.verify_flg = args.verify

        if not self.verify_flg and not args.workerThreads:
            logger.warning('\nNo value specified for --workerThreads. Using the default option of {}'.format(
                self.thread_num))
            logger.info(
                'Important - You can configure the thread count setting using the --workerThreads argument to maximize your download speed.\n')
        # for copying files directly to another s3 bucket
        self.custom_user_s3_endpoint = args.s3_destination

        # non-configurable default instance variables
        self.download_queue = Queue()
        # self.download_queue_metadata = {}  # map of package-file-id to alias
        self.package_file_download_errors = set()
        # self.package_file_download_errors needs a lock if multiple threads will be adding to it simultaneously
        self.package_file_download_errors_lock = threading.Lock()

        self.download_job_uuid = None

        self.download_job_manifest_column_defs = {
            'uuid': self.download_job_uuid,
            'run_date': time.strftime("%Y%m%dT%H%M%S"),
            'package_id': self.package_id,
            'download_directory': self.download_directory,
            's3_destination': self.custom_user_s3_endpoint,
            'data_structure': self.data_structure,
            's3_links_file': self.s3_links_file,
            'regex': self.regex_file_filter
        }
        self.download_job_progress_report_column_defs = {
            'package_file_id': None,
            'package_file_expected_location': None,
            'nda_s3_url': None,
            'exists': False,
            'expected_file_size': None,
            'actual_file_size': 0,
            'e_tag': None,
            'download_complete_time': None
        }
        self.download_progress_report_file_path = self.initialize_verification_files()
        self.default_download_batch_size = 50
        self.metadata_file_path = os.path.join(self.package_metadata_directory, NDATools.NDA_TOOLS_PACKAGE_FILE_METADATA_TEMPLATE % self.package_id)

    # exlcude arg list is the long-parameter name
    def build_rerun_download_cmd(self, exclude_arg_list):
        download_cmd = 'downloadcmd -dp {}'.format(self.package_id)

        if self.download_mode == 'text' and '--txt' not in exclude_arg_list:
            download_cmd += ' -t {}'.format(self.s3_links_file)
        if self.download_mode == 'datastructure' and '--datastructure' not in exclude_arg_list:
            download_cmd += ' -ds {}'.format(self.data_structure)
        if self.regex_file_filter and '--file-regex' not in exclude_arg_list:
            download_cmd += ' --file-regex {}'.format(self.regex_file_filter)
        if self.username and '--username' not in exclude_arg_list:
            download_cmd += ' -u {}'.format(self.username)
        if self.download_directory and '--directory' not in exclude_arg_list:
            download_cmd += ' -d {}'.format(self.download_directory)
        if self.verify_flg and '--verify' not in exclude_arg_list:
            download_cmd += ' --verify'
        if self.thread_num and '--workerThreads' not in exclude_arg_list:
            download_cmd += ' -wt {}'.format(self.thread_num)
        if self.custom_user_s3_endpoint and '--s3-destination' not in exclude_arg_list:
            download_cmd += ' -s3 {}'.format(self.custom_user_s3_endpoint)

        return download_cmd

    def start(self):
        logger.info('')
        logger.info('Getting Package Information...')
        package_resource = self.get_package_info()
        logger.info('')
        logger.info('Package-id: {}'.format(self.package_id))
        logger.info('Name: {}'.format(package_resource['description']))
        logger.info('Has associated files?: {}'.format('Yes' if package_resource['has_associated_files'] else 'No'))
        # Dont print this out because at the moment, the number coming back from the service includes duplicates.
        # uncomment when that is fixed
        print('Number of files in package: {}'.format(package_resource['file_count']))
        logger.info('Total Package Size: {}'.format(Utils.human_size(package_resource['total_package_size'])))
        logger.info('')

        # self.save_package_file_metadata()
        logger.debug('downloading package metadata-file')
        self.download_package_metadata_file()
        df = []
        if self.download_mode == 'datastructure':
            logger.info('Downloading S3 links from data structure: {}'.format(self.data_structure))
            if not package_resource['has_associated_files']:
                logger.info(''''No Associated files detected in this package. In order to download associated files, you must create a new package
        on the NDA website and make sure that you check the option to "Include associated files"''')
                exit_error()
            df = self.use_data_structure()
        elif self.download_mode == 'text':
            logger.info('Downloading S3 links from text file: {}'.format(self.s3_links_file))
            df = self.use_s3_links_file()
        elif self.download_mode == 'package':
            df = self.get_all_files_in_package()
        else:
            df = self.query_files_by_s3_path(self.inline_s3_links)

        if self.regex_file_filter:
            df = df[df.download_alias.str.contains(self.regex_file_filter)]

        logger.info('')

        success_files = set()
        download_request_count = 0
        download_start_date = datetime.datetime.now()

        download_progress_report = open(self.download_progress_report_file_path, 'a', newline='')
        download_progress_report_writer = csv.DictWriter(download_progress_report,
                                                         fieldnames=self.download_job_progress_report_column_defs, extrasaction='ignore')
        failed_s3_links_file = tempfile.NamedTemporaryFile(mode='a',
                                                           delete=False,
                                                           prefix='failed_s3_links_file_{}'.format(time.strftime("%Y%m%dT%H%M%S")),
                                                           suffix='.csv',
                                                           dir=NDATools.NDA_TOOLS_DOWNLOADCMD_LOGS_FOLDER
                                                           )

        message = 'S3 links for files that failed to download will be written out to {}. You can attempt to download these files later by running: ' \
            .format(failed_s3_links_file.name)
        message += '\n\t{} -t "{}"' \
            .format(self.build_rerun_download_cmd(['--text', '--datastructure']), failed_s3_links_file.name)
        logger.info(message)
        logger.info('')
        #time.sleep(1.5)

        file_ct =  df['download_alias'].unique().size
        file_sz = df['file_size'].sum()
        tmp = {}

        # remove files that have already been completed
        if self.download_directory:
            tmp = self.get_completed_files_in_user_specified_dir(df)
        if not self.download_directory:
            completed_files = self.get_completed_files_in_download()
            tmp = {int(f['package_file_id']): int(f['actual_file_size']) for f in completed_files}

        completed_file_sz = sum(tmp.values())
        completed_file_ids = set(tmp.keys())
        completed_file_ct = len(completed_file_ids)
        completed_files = tmp = None #remove large structures from memory
        skipping_message = ''

        if completed_file_ct > 0:
            if self.download_directory:
                skipping_message = 'Skipping {} files ({}) which have already been downloaded in {}\n'.format(
                    completed_file_ct, Utils.human_size(completed_file_sz), self.download_directory)
            else:
                download_progress_report_path = os.path.join(self.package_metadata_directory,
                                                             '.download-progress', self.download_job_uuid,
                                                             'download-progress-report.csv')
                skipping_message = 'Skipping {} files ({}) which have already been downloaded according to log file {}.\n'.format(
                    completed_file_ct, Utils.human_size(completed_file_sz), download_progress_report_path)
            file_ct -= completed_file_ct
            file_sz -= completed_file_sz

        if file_ct <= 0:
            if self.regex_file_filter:
                logger.info('No file is found to match the regex patter {}'.format(self.regex_file_filter))
            else:
                logger.info('All files have been downloaded')
            logger.info('')
            logger.info('Exiting Program...')
            return

        message = '{}Beginning download of {}{} files ({}){} to {} using {} threads'.format(
            skipping_message,
            'the remaining ' if skipping_message else '',
            file_ct,
            Utils.human_size(file_sz),
            f' matching {self.regex_file_filter}' if self.regex_file_filter else '',
            self.custom_user_s3_endpoint or self.download_directory,
            self.thread_num)

        logger.info('')
        logger.info(message)
        #time.sleep(5)

        # These are all arrays just so that the print_download_progress_report method can update the variables inside them
        trailing_50_file_bytes = []
        trailing_50_timestamp = [datetime.datetime.now()]

        download_progress_flush_date=[datetime.datetime.now()]

        def write_to_download_progress_report_file(download_record):
            # if file-size =0, there could have been an error. Dont add to file
            newRecord = vars(download_record)
            if type(newRecord['actual_file_size']) is tuple:
                check = newRecord['actual_file_size'][0] > 0
            else:
                check = newRecord['actual_file_size'] > 0
            if check:
                download_progress_report_writer.writerow(newRecord)
                if (datetime.datetime.now()-download_progress_flush_date[0]).seconds>10:
                    download_progress_report.flush()
                    download_progress_flush_date[0]=datetime.datetime.now()
                else:
                    pass

        def print_download_progress_report(num_downloaded):

            byte_total = sum(trailing_50_file_bytes)
            download_progress_message = 'Download Progress Report [{}]: \n    {}/{} queued files downloaded so far. ' \
                .format(datetime.datetime.now().strftime('%b %d %Y %H:%M:%S'), num_downloaded, download_request_count)
            download_progress_message += '\n    Last 50 files contained ~ {} bytes and finished in {} (Hours:Minutes:Seconds). ' \
                .format(Utils.human_size(byte_total),
                        str(datetime.datetime.now() - trailing_50_timestamp[0]).split('.')[0])

            seconds_last_50_files = (datetime.datetime.now() - trailing_50_timestamp[0]).seconds
            if seconds_last_50_files == 0:
                seconds_last_50_files = 1  # avoid a 'division by 0' error

            # convert download speed to bits per second
            avg_speed_bps = Utils.human_size((8 * byte_total) // seconds_last_50_files)
            if avg_speed_bps[-1:] == 'B':
                avg_speed_bps = avg_speed_bps.replace('B', 'bps')
            else:
                avg_speed_bps = avg_speed_bps.replace('bytes', 'bps')

            download_progress_message += '\n    Avg download rate (in bits per second) for the last 50 files is ~ {}.' \
                .format(avg_speed_bps)

            download_progress_message += '\n    Download has been in progress for {} (Hours:Minutes:Seconds).\n' \
                .format(str(datetime.datetime.now() - download_start_date).split('.')[0])

            download_progress_message = '\n' + download_progress_message + '\n'
            logger.info(download_progress_message)
            trailing_50_file_bytes.clear()
            trailing_50_timestamp[0] = datetime.datetime.now()

        def download(package_file, temp_credentials=None):
            # check if  these exist, and if not, get and set:
            download_record = self.download_from_s3link(package_file, temp_credentials, failed_s3_links_file=failed_s3_links_file)
            # dont add bytes if file-existed and didnt need to be downloaded
            if download_record.download_complete_time:
                trailing_50_file_bytes.append(download_record.actual_file_size)
            success_files.add(package_file['package_file_id'])
            num_downloaded = len(success_files)

            if num_downloaded % 50 == 0:
                print_download_progress_report(num_downloaded)

            download_progress_file_writer_pool.map(write_to_download_progress_report_file, [[download_record]])

        download_pool = ThreadPool(self.thread_num, self.thread_num*6)
        download_progress_file_writer_pool = ThreadPool(1, 1000)

        for package_files in self.generate_download_batch_file_ids(completed_file_ids, df):
            if len(package_files) > 0:
                additional_file_ct = len(package_files)
                download_request_count += additional_file_ct
                logger.info('Adding {} files to download queue. Queue contains {} files\n'.format(additional_file_ct,
                                                                                                  download_request_count))
                pkfiles = {f['package_file_id']:f for f in package_files}
                if self.custom_user_s3_endpoint:
                    # we dont need presigned urls if the user is transferring to their own s3 bucket
                    file_id_to_cred_list={f['package_file_id']:None for f in package_files}
                else:
                    file_id_to_cred_list = self.get_presigned_urls(list(pkfiles.keys()))
                download_pool.map(download, [ [pkfiles[file_id], file_credentials] for file_id, file_credentials in file_id_to_cred_list.items()])

        download_pool.wait_completion()
        download_progress_file_writer_pool.wait_completion()
        failed_s3_links_file.flush()
        failed_s3_links_file.close()
        download_progress_report.flush()
        download_progress_report.close()

        # dont generate a file if there were no failures
        if not self.package_file_download_errors:
            logger.info('No failures detected. Removing file {}'.format(failed_s3_links_file.name))
            os.remove(failed_s3_links_file.name)

        logger.info('')

        logger.info('Finished processing all download requests @ {}.'.format(datetime.datetime.now()))
        logger.info('     Total download requests {}'
                    .format(download_request_count))

        download_error_count = len(self.package_file_download_errors)
        logger.info('     Total errors encountered: {}'.format(download_error_count))

        if download_error_count > 0:
            logger.info('     Failed to download {} files. See {} for more details'.format(download_error_count, failed_s3_links_file.name))

        logger.info('')
        logger.info(' Exiting Program...')

    def download_local(self, download_request, err_if_exists=False):
        #completed_download = os.path.normpath(os.path.join(self.download_directory, download_request.package_file_relative_path))
        downloaded = False
        resume_header = None

        # check if we are downloading from alt endpoint where bucket name contains dots.
        def get_http_adapter(s3_link):
            bucket, path = deconstruct_s3_url(s3_link)
            config = {'max_retries': 10}
            if ('.' in bucket):
                return AltEndpointSSLAdapter(**config)
            return HTTPAdapter(**config)

        # use this instead of exists_ok in order to work with python v.2
        def mk_dir_ignore_err(dir):
            try:
                os.makedirs(os.path.normpath(dir))
            except FileExistsError as e:
                pass
            except OSError as e:
                # Raise exception for any errors other than FileExists error
                # Using OSError for version compatibility
                if e.errno != 17:
                    raise
                pass
        if os.path.isfile(download_request.completed_download_abs_path):
            if err_if_exists:
                msg = "File {} already exists. Move or rename the file before re-running the command to continue".format(
                    download_request.completed_download_abs_path)
                logger.info(msg)
                logger.info('Exiting...')
                sys.exit(1)

            logger.info('Skipping download (already exists): {}'.format(download_request.completed_download_abs_path))
            actual_size = os.path.getsize(download_request.completed_download_abs_path)
            download_request.actual_file_size = actual_size
            download_request.exists = True
            return download_request

        if os.path.isfile(download_request.partial_download_abs_path):
            downloaded = True
            downloaded_size = os.path.getsize(download_request.partial_download_abs_path)
            resume_header = {'Range': 'bytes={}-'.format(downloaded_size)}
            logger.info('Resuming download: {}'.
                        format(download_request.partial_download_abs_path))
        else:
            mk_dir_ignore_err(os.path.dirname(download_request.partial_download_abs_path))
            logger.info('Starting download: {}'.format(download_request.partial_download_abs_path))
            # downloading to local machine
        bytes_written = 0
        with requests.session() as s:
            s.mount(download_request.presigned_url, get_http_adapter(download_request.presigned_url))
            if resume_header:
                s.headers.update(resume_header)
            with open(download_request.partial_download_abs_path, "ab" if downloaded else "wb") as download_file:
                with s.get(download_request.presigned_url, stream=True) as response:
                    response.raise_for_status()
                    for chunk in response.iter_content(chunk_size=1024 * 1024 * 5):  # iterate 5MB chunks
                        if chunk:
                            bytes_written += download_file.write(chunk)
        # TODO - this doesnt work when using s3fs...add ticket to make it easy to download using s3fs
        os.rename(download_request.partial_download_abs_path, download_request.completed_download_abs_path)
        logger.info('Completed download {}'.format(download_request.completed_download_abs_path))
        download_request.actual_file_size = bytes_written
        bucket, key = Utils.deconstruct_s3_url(download_request.presigned_url)
        download_request.nda_s3_url = 's3://{}/{}'.format(bucket, key)

    def download_to_s3(self, download_request):
        # downloading directly to s3 bucket
        # get cred for file
        response = self.get_temp_creds_for_file(download_request.package_file_id, self.custom_user_s3_endpoint)
        ak = response['access_key']
        sk = response['secret_key']
        sess_token = response['session_token']
        source_uri = response['source_uri']
        dest_uri = response['destination_uri']

        download_request.nda_s3_url = source_uri

        dest_bucket, dest_path = Utils.deconstruct_s3_url(dest_uri)
        src_bucket, src_path = Utils.deconstruct_s3_url(source_uri)

        logger.info('Starting download: s3://{}/{}'.format(dest_bucket, dest_path))

        # boto3 copy
        sess = boto3.session.Session(aws_access_key_id=ak,
                                     aws_secret_access_key=sk,
                                     aws_session_token=sess_token,
                                     region_name='us-east-1')

        s3_client = sess.client('s3')
        response = s3_client.head_object(Bucket=src_bucket, Key=src_path)
        download_request.actual_file_size = response['ContentLength']
        download_request.e_tag = response['ETag'].replace('"', '')

        s3 = sess.resource('s3')
        copy_source = {
            'Bucket': src_bucket,
            'Key': src_path
        }

        def print_upload_part_info(bytes):
            logger.info('Transferred {} for {}'.format(Utils.human_size(bytes), download_request.nda_s3_url))

        KB = 1024
        MB = KB * KB
        GB = KB ** 3
        LARGE_OBJECT_THRESHOLD = 5 * GB
        args = {
            'ExtraArgs': {'ACL': 'bucket-owner-full-control'}
        }

        if int(download_request.actual_file_size) >= LARGE_OBJECT_THRESHOLD:
            logger.info('Transferring large object {} ({}) in multiple parts'
                        .format(download_request.nda_s3_url,
                                Utils.human_size(int(download_request.actual_file_size))))
            config = TransferConfig(multipart_threshold=LARGE_OBJECT_THRESHOLD, multipart_chunksize=1 * GB)
            args['Config'] = config
            args['Callback'] = print_upload_part_info

        s3.meta.client.copy(copy_source,
                            dest_bucket,
                            dest_path,
                            **args)

    def handle_download_exception(self, download_request, e, download_local, err_if_exists, package_file, failed_s3_links_file=None):
        if not download_request.presigned_url and download_local:
            # we couldnt get credentials, which means the service has become un-responsive.
            # Instruct the user to retry at another time
            logger.info('')
            logger.info(
                'Unexpected Error During File Download - Service Unresponsive. Unable to obtain credentials for file-id {}'.format(
                    download_request.package_file_id))
            logger.info('Please re-try downloading files at a later time. ')
            logger.info('You may contact NDAHelp@mail.nih.gov for assistance in resolving this error.')
            # use os._exit to kill the whole program. This works even if this is called in a child thread, unlike sys.exit()
            os._exit(1)

        self.write_to_failed_download_link_file(failed_s3_links_file, s3_link=download_request.presigned_url, source_uri=download_request.nda_s3_url)

        error_code = -1 if not isinstance(e, HTTPError) else int(e.response.status_code)
        if error_code == 404:
            message = 'This path is incorrect: {}. Please try again.'.format(download_request.presigned_url)
            logger.error(message)
        elif error_code == 403:
            #if we are using expired credentials, regenerate and resume the download
            credentials_are_expired = False
            if download_local:
                if isinstance(e, requests.exceptions.HTTPError) and 'Request has expired' in e.response.text:
                    credentials_are_expired = True
                    
            if credentials_are_expired:
                logger.warning(f'Temporary credentials have expired for file {download_request.package_file_id}. Regenerating credentials and restarting download')
                presigned_url = self.get_temp_creds_for_file(download_request.package_file_id)
                return self.download_from_s3link(package_file, presigned_url, download_local, err_if_exists, failed_s3_links_file )
            else:
                message = '\nThis is a private bucket. Please contact NDAR for help: {}'.format(download_request.presigned_url)
                logger.error(message)
        else:
            logger.error(str(e))
            logger.error(get_traceback())
            if 'operation: Access Denied' in str(e):
                logger.error('')
                logger.error(
                    'This error is likely caused by a misconfiguration on the target s3 bucket')
                logger.error(
                    "For more information about how to correctly configure the target bucket, run 'downloadcmd -h' and read the description of the s3 argument")
                logger.error('')
                time.sleep(2)

        # if source_uri is set, it means they're downloading to s3 bucket and there will not be any partial file
        if download_request.actual_file_size == 0 and not download_local:
            try:
                os.remove(download_request.partial_download_abs_path)
            except:
                logger.error('error removing partial file {}'.format(download_request.partial_download_abs_path))

    def download_from_s3link(self, package_file, presigned_url, download_local=None, err_if_exists=False, failed_s3_links_file=None, download_dir=None):
        if download_local is None:
            download_local = False if self.custom_user_s3_endpoint else True
        if not download_dir:
            download_dir = self.download_directory

        download_request = DownloadRequest(package_file, presigned_url, self.package_id, download_dir)
        try:
            if download_local:
                self.download_local(download_request, err_if_exists)
            else:
                self.download_to_s3(download_request)
            download_request.exists = True
            download_request.download_complete_time = time.strftime("%Y%m%dT%H%M%S")
            return download_request
        except Exception as e:
            self.handle_download_exception(download_request, e, download_local, err_if_exists,package_file, failed_s3_links_file)
            return download_request

    def write_to_failed_download_link_file(self, failed_s3_links_file, s3_link, source_uri):
        src_bucket, src_path = Utils.deconstruct_s3_url(s3_link if s3_link else source_uri)
        s3_address = 's3://' + src_bucket + '/' + src_path

        with self.package_file_download_errors_lock:
            self.package_file_download_errors.add(s3_address)
            if failed_s3_links_file:
                failed_s3_links_file.write(s3_address + "\n")
                failed_s3_links_file.flush()

    def generate_download_batch_file_ids(self, completed_file_ids, df):
        batch = []
        size = 0
        for _, row in df.iterrows():
            if row['package_file_id'] not in completed_file_ids:
                batch.append(row)
                size += 1
                if size % self.default_download_batch_size == 0:
                    yield batch
                    batch = []
                    size = 0
        yield batch

    def find_matching_download_job(self, download_job_manifest_path):
        def is_job_match(possible_match):
            must_match = [
                'data_structure',
                'download_directory',
                's3_destination',
                's3_links_file',
                'package_id',
                'regex'
            ]

            def test_match(key):
                # None gets converted to empty string.
                # Convert it back if empty string is detected so that None values can be compared using == operator
                val1 = possible_match[key] or None
                # values from download_job_manifest_column_defs will never be None, instead they will be an empty string ''
                val2 = self.download_job_manifest_column_defs[key]
                if key == 'download_directory':
                    val2 = Utils.convert_to_abs_path(val2)
                    val1 = Utils.convert_to_abs_path(val1)
                elif key == 's3_links_file':
                    # only convert to basename if values are specified for both
                    if val2 and val1:
                        val2 = os.path.basename(val2)
                        val1 = os.path.basename(val1)
                elif key == 'package_id':
                    if val2 and val1:
                        val2 = int(val2)
                        val1 = int(val1)
                return val1 == val2

            return all(map(test_match, must_match))

        # FIND MATCHING JOB RECORD IF IT EXISTS
        with open(download_job_manifest_path, newline='') as csvfile:
            job_reader = csv.DictReader(csvfile)
            for job in job_reader:
                if is_job_match(job):
                    return job
        return None

    def initialize_verification_files(self):

        download_job_manifest_columns = self.download_job_manifest_column_defs.keys()

        def initialize_download_progress_file(fp):
            with open(fp, 'a', newline='') as file:
                writer = csv.DictWriter(file, fieldnames=self.download_job_progress_report_column_defs)
                writer.writeheader()

        def add_entry_to_job_manifest(fp):
            self.download_job_uuid = self.download_job_manifest_column_defs['uuid'] = str(uuid.uuid4())
            with open(fp, 'a', newline='') as file:
                writer = csv.DictWriter(file, fieldnames=download_job_manifest_columns)
                writer.writerow(self.download_job_manifest_column_defs)

        def initialize_job_manifest_file(fp):
            with open(fp, 'w', newline='') as file:
                writer = csv.DictWriter(file, fieldnames=download_job_manifest_columns)
                writer.writeheader()

        if not os.path.exists(self.package_metadata_directory):
            os.mkdir(self.package_metadata_directory)

        DOWNLOAD_PROGRESS_FOLDER = os.path.join(self.package_metadata_directory, '.download-progress')
        if not os.path.exists(DOWNLOAD_PROGRESS_FOLDER):
            os.mkdir(DOWNLOAD_PROGRESS_FOLDER)

        download_job_manifest_path = os.path.join(DOWNLOAD_PROGRESS_FOLDER, 'download-job-manifest.csv')
        if not os.path.exists(download_job_manifest_path):
            initialize_job_manifest_file(download_job_manifest_path)

        job_record = self.find_matching_download_job(download_job_manifest_path)
        if job_record is not None:
            self.download_job_uuid = job_record['uuid']

        if not self.download_job_uuid:
            add_entry_to_job_manifest(download_job_manifest_path)

        DOWNLOAD_JOB_UUID_DIR = os.path.join(DOWNLOAD_PROGRESS_FOLDER, str(self.download_job_uuid))
        if not os.path.exists(DOWNLOAD_JOB_UUID_DIR):
            os.mkdir(DOWNLOAD_JOB_UUID_DIR)

        download_progress_report_file = os.path.join(DOWNLOAD_JOB_UUID_DIR, 'download-progress-report.csv')
        if not os.path.exists(download_progress_report_file):
            initialize_download_progress_file(download_progress_report_file)
        return download_progress_report_file

    '''        
            Steps -
            1. Find path of correct download-progress-report file from .download-progress/download-job-manifest.csv (if it exists),
                for the given target directory and download mode
            2. Make a copy of the download-progress-report and save into verification folder. Name it 'download-verification-report.csv'
            3. Consider everything currently in the download-verification-report.csv where expected file-size=actual file-size as being downloaded
                     a. there really shouldnt be any entries where the expected filesize doesnt match actual size , but run the code anyway
            4. Read file-names of download-verification-report.csv into set() in memory
            *5. Get the complete file-listing for the download (using the provided arguments -d, -ds, -t and -dp)
            6. Add anything that is not in the set in step 4 into the download-verification-report.csv
            7. Run os.path.exists() on each added entry - if it exists, add file-size info to download-verification-report.csv. if not, set
            exists columns and file-size columns
            8. Run through the file download-verification-report.csv and collect everything where the actual and expected file-sizes dont match up.
            9. Create the download-verification-retry-s3-links.csv file from the entries found in step 8
    '''

    def verify_download(self):
        self.download_package_metadata_file()

        if self.custom_user_s3_endpoint:
            raise Exception(
                'The --verify command does not yet support checking for files in s3 endpoints. This feature will be added in a future iteration...')
            exit_error()

        verification_report_path = os.path.join(self.package_metadata_directory,
                                                'download-verification-report.csv')
        err_mess_template = 'Cannot start verification process - {} already exists \nYou must move or rename the file in order to continue'
        if os.path.exists(verification_report_path):
            logger.info('')
            logger.info(err_mess_template.format(verification_report_path))
            exit_error()

        fpath = os.path.join(self.package_metadata_directory, 'download-verification-retry-s3-links.csv')
        if os.path.exists(fpath):
            logger.info(err_mess_template.format(fpath))
            exit_error()

        def get_download_progress_report_path():
            return os.path.join(self.package_metadata_directory, '.download-progress',
                                self.download_job_uuid, 'download-progress-report.csv')

        def parse_download_progress_report_for_files(download_progress_report_path):
            files = []
            if os.path.exists(download_progress_report_path):
                with open(download_progress_report_path, newline='') as csvfile:
                    file_reader = csv.DictReader(csvfile)
                    files = [f for f in file_reader]
            return files

        def get_complete_file_list():
            if self.download_mode in ['text', 'datastructure']:
                if self.download_mode == 'datastructure':
                    df = self.use_data_structure()
                elif self.download_mode == 'text':
                    df = self.use_s3_links_file()
                return df
            elif self.download_mode == 'package':
                df = self.get_all_files_in_package()
                return df
            else:
                raise Exception('Unsupported download mode: {}'.format(self.download_mode))

        def create_download_verification_retry_links_file(s3_links):
            fpath = os.path.join(self.package_metadata_directory,
                                 'download-verification-retry-s3-links.csv')
            with open(fpath, 'w') as retry_file:
                for link in s3_links:
                    retry_file.write(link + '\n')

        def add_files_to_report(download_progress_report_path, verification_report_path, probably_missing_files_list, df):
            copyfile(download_progress_report_path, verification_report_path)

            all_records = []
            for file_id in probably_missing_files_list:
                file_info = df[df['package_file_id']==file_id].to_dict('records')[0]
                record = copy.deepcopy(self.download_job_progress_report_column_defs)
                # TODO - consider making these the same names
                record['package_file_expected_location'] = file_info['download_alias']
                record['expected_file_size'] = min(abs(int(file_info['file_size'])),1)
                record['package_file_id'] = int(file_id)
                if file_info['download_alias'] == (pathlib.Path(self.metadata_file_path).name + '.gz'):
                    download_path = os.path.join(self.package_metadata_directory,
                                                 file_info['download_alias'])
                else:
                    download_path = os.path.join(self.download_directory,
                                             file_info['download_alias'])
                if os.path.exists(download_path):
                    record['exists'] = True
                    stat = os.stat(download_path)
                    record['actual_file_size'] = stat.st_size
                    if record['actual_file_size'] == record['expected_file_size']:
                        record['download_complete_time'] = time.strftime("%Y%m%dT%H%M%S", time.localtime(stat.st_ctime))

                all_records.append(record)

            # batch requests to reduce overhead
            # TODO - add s3 location to file resource in order to eliminate this step?
            logger.info('Retrieving s3 url information for {} nda file records'.format(len(all_records)))
            batch_size = 1000
            batches = [all_records[i:i + batch_size] for i in range(0, len(all_records), batch_size)]
            for batch in batches:
                result = self.get_presigned_urls([record['package_file_id'] for record in batch])
                # result is a dictionary of package-file-id to presignedUrl
                for record in batch:
                    ps_url = result[record['package_file_id']]
                    dest_bucket, dest_path = Utils.deconstruct_s3_url(ps_url)
                    record['nda_s3_url'] = 's3://{}/{}'.format(dest_bucket, dest_path)

            with open(verification_report_path, 'a', newline='') as verification_report:
                download_progress_report_writer = csv.DictWriter(verification_report,
                                                                 fieldnames=self.download_job_progress_report_column_defs)
                download_progress_report_writer.writerows(all_records)

            return [record['nda_s3_url'] for record in all_records if
                    record['actual_file_size'] < record['expected_file_size']]

        logger.info('')
        logger.info(
            'Running verification process. This process will check whether all of the files from the following downloadcmd were successfully downloaded to the computer:')

        verification_report_path = os.path.join(self.package_metadata_directory,
                                                'download-verification-report.csv')

        logger.info('{}'.format(self.build_rerun_download_cmd(['--verify'])))
        logger.info('')
        pr_path = get_download_progress_report_path()
        logger.info('Getting expected file list for download...')
        df = get_complete_file_list()
        df = df.rename(columns={c:c.lower() for c in df.columns})
        complete_file_set = set(df['package_file_id'].values)
        # Sometimes there are dupes in the qft table. eliminate to get accurate file count

        accurate_file_ct = df['download_alias'].unique().size
        file_sz = Utils.human_size(df['file_size'].sum())

        logger.info(
            '{} files are expected to have been downloaded from the command above, totaling {}'.format(accurate_file_ct,
                                                                                                       file_sz))
        logger.info('')
        logger.info('Parsing program system logs for history of completed downloads...')
        logger.info(
            'Important - if you think files may have been deleted from your system after the download was run, you should remove the system log at {}'
            ' and re-run the --verify command. This will force the program to check for these files instead of assuming they exist based on system log entries. This will cause the --verify step to take longer'
            ' to finish but will be necessary for accurate results.'.format(pr_path))
        downloaded_file_records = parse_download_progress_report_for_files(pr_path)
        # There shouldn't be duplicates in the system logs, but check anyway
        downloaded_file_records_count = len({f['package_file_expected_location'] for f in downloaded_file_records})
        downloaded_file_set = {int(f['package_file_id']) for f in downloaded_file_records}
        logger.info('')
        logger.info(
            'Found {} complete file downloads according to log file {}'.format(downloaded_file_records_count, pr_path))
        logger.info('')
        probably_missing_files = complete_file_set - downloaded_file_set
        logger.info(
            'Checking {} for all files which were not found in the program system logs. Detailed report will be created at {}...'
            .format(self.download_directory, verification_report_path))
        undownloaded_s3_links = add_files_to_report(pr_path, verification_report_path, probably_missing_files, df)
        logger.info('')
        if undownloaded_s3_links:
            logger.info(
                'Finished verification process and file check. Found {} files that were missing or whose size on disk were less than expected'.format(
                    len(undownloaded_s3_links)))
            logger.info('')
            logger.info(
                'Generating list of s3 links for all missing/incomplete files...'.format(len(undownloaded_s3_links)))
            create_download_verification_retry_links_file(undownloaded_s3_links)
            incomplete_s3_fp = os.path.join(self.package_metadata_directory,
                                            'download-verification-retry-s3-links.csv')
            logger.info(
                'Finished creating {} file. \nThis file contains s3-links for all files that were found to be missing or incomplete. You may '
                'download these files by running:\n'
                '   {} -t {}'.format(incomplete_s3_fp,
                                     self.build_rerun_download_cmd(['--verify', '--text', '--datastructure']),
                                     incomplete_s3_fp))
        else:
            logger.info(
                'Finished verification process and file check. No missing files found. All files match expected size. Download 100% complete.')

        logger.info('')
        logger.info(
            'Details about status of files in download can be found at {} (This file can be opened with Excel or Google Spreadsheets)'.format(
                verification_report_path))
        exit_normal()

    def get_temp_creds_for_file(self, package_file_id, custom_user_s3_endpoint=None):
        url = self.package_url + '/{}/files/{}/download_token'.format(self.package_id, package_file_id)
        if custom_user_s3_endpoint:
            s3_dest_bucket, s3_dest_prefix = Utils.deconstruct_s3_url(custom_user_s3_endpoint)
            url += '?s3SourceBucket={}'.format(s3_dest_bucket)
            if s3_dest_prefix:
                url += '&s3SourcePrefix={}'.format(s3_dest_prefix)
        tmp = get_request(url, auth=self.auth,
                          error_handler=HttpErrorHandlingStrategy.reraise_status,
                          deserialize_handler=DeserializeHandler.none)
        return json.loads(tmp.text)

    def generate_metadata_and_get_creds(self):
        logger.info(f'Getting list of all files in package at {time.strftime("%H:%M:%S")} ....')
        print('This is a one time operation that will take about 1 min for every 1 million files in your package.')
        print('Your download will start after this process completes.... ')
        end_time = datetime.datetime.now() + datetime.timedelta(minutes=30)
        self.request_metadata_file_creation()
        while True:
            try:
                creds = self.get_package_file_metadata_creds()
                logger.info(f'List of files retrieved at {time.strftime("%H:%M:%S")}...')
                return creds
            except:
                time.sleep(15)
                if datetime.datetime.now() > end_time:
                    logger.error('Error during creation of package meta-data file')
                    logger.error('\nPlease contact NDAHelp@mail.nih.gov for help in resolving this error')
                    exit_error()

    def download_package_metadata_file(self):
        if os.path.exists(self.metadata_file_path):
            logger.debug('Metadata file exists')
            return
        try:
            creds = self.get_package_file_metadata_creds()
        except:
            creds = self.generate_metadata_and_get_creds()

        file_resource = self.get_package_file(creds['package_file_id'])
        self.download_from_s3link(file_resource, creds['downloadURL'], download_local=True, download_dir=self.package_metadata_directory)
        download_location = f"{self.metadata_file_path}.gz"
        outfile = download_location.rstrip('.gz')
        logger.debug(f'unzipping metadata file at {time.strftime("%H:%M:%S")}...')
        with gzip.open(download_location, 'rb') as f_in:
            with open(outfile, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        return outfile

    def get_package_file_metadata_creds(self):
        url = self.package_url + \
              '/{}/files/package_file_metadata'.format(self.package_id)
        tmp = get_request(url, auth=self.auth,
                          error_handler=HttpErrorHandlingStrategy.reraise_status,
                          deserialize_handler=DeserializeHandler.convert_json)
        return tmp

    def get_package_file(self, file_id):
        url = self.package_url + \
              '/{}/files/{}'.format(self.package_id, file_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.convert_json)
        return tmp

    def get_files_from_datastructure(self, data_structure):
        df = pd.read_csv(self.metadata_file_path, header=0)
        df = self.rename_df_columns_to_lowercase(df)
        return df[df['short_name'] == data_structure]

    def get_data_structure_manifest_file_info(self):
        url = self.package_url + \
              '/{}/files?page=1&size=all&types=Package%20Metadata&regex={}'.format(self.package_id,
                                                                                   'datastructure_manifest.txt')
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        results = json.loads(tmp.text)['results']
        # return None instead of empty list, since this method is always supposed to return 1 thing
        return results[0] if results else None

    def get_data_structure_files(self):
        url = self.package_url + \
              '/{}/files?page=1&size=all&types=Data'.format(self.package_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        tmp.raise_for_status()
        results = json.loads(tmp.text)['results']
        return [r for r in results if r['nda_file_type'] == 'Data']

    def get_data_structure_file_info(self, short_name):
        url = self.package_url + \
              '/{}/files?page=1&size=all&types=Package%20Metadata&types=Data&regex={}'.format(self.package_id,
                                                                                              short_name)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        tmp.raise_for_status()
        results = json.loads(tmp.text)['results']
        # return None instead of empty list, since this method is always supposed to return 1 thing
        return results[0] if results else None

    def get_package_file_info(self, file_id):
        url = self.package_url + '/{}/files/{}'.format(self.package_id, file_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        return json.loads(tmp.text)

    def get_package_info(self):
        url = self.package_url + '/{}'.format(self.package_id)
        tmp = get_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        return json.loads(tmp.text)

    def get_package_files_by_page(self, page, batch_size):
        url = self.package_url + '/{}/files?page={}&size={}'.format(self.package_id, page, batch_size)
        if self.regex_file_filter:
            url += '&regex={}'.format(self.regex_file_filter)
        try:
            tmp = get_request(url, auth=self.auth,
                              error_handler=HttpErrorHandlingStrategy.reraise_status, deserialize_handler=DeserializeHandler.none)
            tmp.raise_for_status()
            response = json.loads(tmp.text)
            return response['results']

        except HTTPError as e:
            if 'Cannot navigate past last page' in e.response.text:
                # empty alias and package-file return means there are no files for given page parameter
                return []
            else:
                raise e
        except Exception as e:
            raise e

    def get_presigned_urls(self, id_list):
        """
        Stores key-value pairs of (key: package_file_id, value: presigned URL)
        :param id_list: List of package file IDs with max size of 50,000
        """
        # Use the batchGeneratePresignedUrls when retrieving multiple files
        if not self.verify_flg:
            logger.debug('Retrieving credentials for {} files'.format(len(id_list)))
        url = self.package_url + '/{}/files/batchGeneratePresignedUrls'.format(self.package_id)
        response = post_request(url,  payload=id_list, auth=self.auth,
                                error_handler=HttpErrorHandlingStrategy.reraise_status, deserialize_handler=DeserializeHandler.convert_json)
        creds = {e['package_file_id']: e['downloadURL'] for e in response['presignedUrls']}
        if not self.verify_flg:
            logger.debug('Finished retrieving credentials')
        return creds

    def get_completed_files_in_download(self):
        download_progress_report_path = os.path.join(self.package_metadata_directory,
                                                     '.download-progress', self.download_job_uuid,
                                                     'download-progress-report.csv')

        files = []
        if os.path.exists(download_progress_report_path):
            with open(download_progress_report_path, newline='') as csvfile:
                file_reader = csv.DictReader(csvfile)
                files = [f for f in file_reader if bool(f['exists'])]
        return files

    def get_completed_files_in_user_specified_dir(self, df):
        downloaded_files = {}

        if not os.path.exists(self.download_directory):
            os.mkdir(self.download_directory)
        if len(os.listdir(self.download_directory)) == 0:
            return downloaded_files

        for (root, dir, file) in os.walk(self.download_directory):
            for f in file:
                downloaded_file_path = os.path.join(root, f)
                if '.partial' in downloaded_file_path:
                    os.remove(downloaded_file_path)
                else:
                    for row in df.index:
                        if os.path.normpath(df['download_alias'][row]) == os.path.normpath(downloaded_file_path[len(self.download_directory) + 1:]):
                            downloaded_files[df['package_file_id'][row]] = os.path.getsize(downloaded_file_path)
                            break
        return downloaded_files

    def get_all_files_in_package(self):
        df = pd.read_csv(self.metadata_file_path, header=0)
        return self.rename_df_columns_to_lowercase(df)

    def use_s3_links_file(self):
        """
        Reads a text file line by line, collects all S3 URLs, and requests the package file ids
        from the package service for those files.
        """
        path_list = set()
        try:
            with open(self.s3_links_file, 'r', encoding='utf-8') as s3_links_file:
                for line in s3_links_file:
                    if line.startswith('s3://'):
                        path_list.add(line.strip())

            if not path_list:
                logger.info(
                    'No valid paths found in s3-links file. If you specified a regular expression, make sure your regular expression is correct before re-running the command.')
                exit_error()

            return self.query_files_by_s3_path(path_list)
        except IOError as e:
            logger.error(
                '{} not found. Please enter the correct path to your file and try again.'.format(self.data_structure))
            raise e

    def use_data_structure(self):
        """
        Downloads a data structure file from the package, the data_structure_manifest file,
        parses both files to identify all associated files for the data structure, and then
        requests the package file ids for those files from the package service
        """
        df = self.get_files_from_datastructure(self.data_structure)
        if df.empty:
            logger.info(
                '{} data structure is not included in the package {}'.format(self.data_structure, self.package_id))
            exit_error()
        return df

    def rename_df_columns_to_lowercase(self, df):
        return df.rename (columns={c:c.lower() for c in df.columns})

    def query_files_by_s3_path(self, path_list):
        if not path_list:
            exit_error(message='Illegal Argument - path_list cannot be empty')
        df = pd.read_csv(self.metadata_file_path, header=0)
        df = self.rename_df_columns_to_lowercase(df)
        return df[df['nda_s3_url'].isin(path_list)]

    def request_metadata_file_creation(self):
        url = self.package_creation_url + \
              '/{}/create-package-metadata-file'.format(self.package_id)
        tmp = post_request(url, auth=self.auth, deserialize_handler=DeserializeHandler.none)
        return tmp




from __future__ import absolute_import, with_statement

import concurrent
import copy
import datetime
import math
import multiprocessing
import queue
from pathlib import Path

import requests.exceptions
from boto3.s3.transfer import TransferConfig
from s3transfer.constants import GB
from tqdm import tqdm

from NDATools.Configuration import *
from NDATools.Utils import get_request, put_request, DeserializeHandler, post_request, get_s3_client_with_config

logger = logging.getLogger(__name__)


class Status:
    UPLOADING = 'Uploading'
    SYSERROR = 'SystemError'
    COMPLETE = 'Complete'
    ERROR = 'Error'
    PROCESSING = 'In Progress'
    READY = 'Ready'


class LocalAssociatedFile:
    sys_path: Path
    size_bytes: int
    csv_value: str

    def __init__(self, path, size_bytes, csv_value, file_resource):
        self.sys_path = path
        self.size_bytes = size_bytes
        self.csv_value = csv_value
        self.file_resource = file_resource

    def to_api_resource(self):
        tmp = copy.deepcopy(self.file_resource)
        tmp['size'] = self.size_bytes
        return tmp


def to_local_associated_file(file_resource, directory_search_list) -> LocalAssociatedFile:
    # TODO consider adding cwd to directory_search_list by default
    csv_path = file_resource['file_user_path']
    for directory in directory_search_list:
        try:
            sys_path = Path(directory, csv_path)
            size = os.path.getsize(sys_path)
            return LocalAssociatedFile(sys_path, size, csv_path, file_resource)
        except OSError:
            pass
    logger.debug(f'file {csv_path} not found in any of the following directories: {",".join(directory_search_list)}')
    return None


class Submission:
    def __init__(self, config, submission_id=None, package_id=None, allow_exit=False, username=None, password=None,
                 thread_num=None, batch_size=None):
        assert submission_id or package_id, "Either submission-id or package-id must be specified"
        self.config = config
        self.api = self.config.submission_api
        if username:
            self.config.username = username
        if password:
            self.config.password = password
        self.username = self.config.username
        self.password = self.config.password
        self.__files = []
        self.upload_queue = queue.Queue()
        self.thread_num = max([1, multiprocessing.cpu_count() - 1])
        if thread_num:
            self.thread_num = thread_num
        self.batch_size = 10000
        if batch_size:
            self.batch_size = batch_size
        self.credentials_list = []
        self.status = None
        self.source_bucket = None
        self.upload_tries = 0
        self.max_submit_time = 120
        self.auth = requests.auth.HTTPBasicAuth(self.config.username, self.config.password)
        self.submission_id = submission_id
        self.package_id = package_id
        self.no_read_access = set()
        self.exit = allow_exit
        self.all_mpus = []
        if not self.config.directory_list:
            self.config.directory_list = [os.getcwd()]

    def get_submission_versions(self):
        return get_request("/".join([self.api, self.submission_id, 'change-history']), auth=self.auth)

    def replace_submission(self):
        version_count = len(self.get_submission_versions())
        put_request(
            "/".join([self.api, self.submission_id]) + "?submissionPackageUuid={}&async=true".format(self.package_id),
            auth=self.auth, deserialize_handler=DeserializeHandler.none)

        # poll the versions endpoint until a new one is created or until we timeout
        end_time = datetime.timedelta(seconds=self.config.validation_timeout) + datetime.datetime.now()
        while True:
            if datetime.datetime.now() > end_time:
                logger.error("Timed out waiting for submission to replace.")
                logger.error('\nPlease email NDAHelp@mail.nih.gov for help in resolving this error')
                exit_error()
            new_version_count = len(self.get_submission_versions())
            if new_version_count > version_count:
                logger.info("Submission replaced successfully.")
                break
            time.sleep(10)

    def query_submissions_by_package_id(self, package_id):
        return get_request(f"{self.api}?packageUuid={package_id}", auth=self.auth)

    def submit(self):
        post_request("/".join([self.api, self.package_id]) + "?async=true", auth=self.auth,
                     deserialize_handler=DeserializeHandler.none)
        # poll the versions endpoint until a new one is created or until we timeout
        end_time = datetime.timedelta(seconds=self.config.validation_timeout) + datetime.datetime.now()
        while True:
            if datetime.datetime.now() > end_time:
                logger.error("Timed out waiting for submission to get created.")
                logger.error('\nPlease email NDAHelp@mail.nih.gov for help in resolving this error')
                exit_error()
            response = self.query_submissions_by_package_id(self.package_id)
            if response:
                logger.debug("Submission successfully created.")
                sub_id = response[0]['submission_id']
                break
            time.sleep(10)
        response = get_request("/".join([self.api, sub_id]), auth=self.auth)
        self.status = response['submission_status']
        self.submission_id = response['submission_id']

    def check_status(self):
        response = get_request("/".join([self.api, self.submission_id]), auth=self.auth)
        self.status = response['submission_status']

    def get_upload_progress(self):
        response = get_request("/".join([self.api, self.submission_id, "upload-progress"]), auth=self.auth)
        return int(response['associated_file_count']), int(response['uploaded_file_count'])

    def get_multipart_credentials(self, file_ids):
        all_credentials = []
        batched_ids = [file_ids[i:i + self.batch_size] for i in range(0, len(file_ids), self.batch_size)]

        for ids in batched_ids:
            query_params = ''
            if self.config.source_bucket is not None:
                query_params = '?s3SourceBucket={}'.format(self.config.source_bucket)
                query_params += '&s3Prefix={}'.format(
                    self.config.source_prefix) if self.config.source_prefix is not None else ''
            credentials_list = post_request("/".join(
                [self.api, self.submission_id, 'files/batchMultipartUploadCredentials']) + query_params,
                                            payload=json.dumps(ids), auth=self.auth)
            all_credentials = all_credentials + credentials_list['credentials']

        return all_credentials

    def get_files_from_page(self, page_number, page_size, exclude_uploaded=False):
        excluded_q_param = f'&omitCompleted=true' if exclude_uploaded else ''
        try:
            get_files_url = "/".join([self.api, self.submission_id,
                                      f'file-listing?pageNumber={page_number}&pageSize={page_size}{excluded_q_param}'])
            response = get_request(get_files_url, auth=self.auth, error_handler=HttpErrorHandlingStrategy.ignore,
                                   deserialize_handler=DeserializeHandler.none)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 400 and 'Cannot navigate past last page' in error.response.text:
                # we got passed the last page
                return []
            else:
                logger.error(error.response.text)
                logger.error('\nPlease email NDAHelp@mail.nih.gov for help in resolving this error')
                exit_error()

    def get_files_iterator(self, total_results, page_size=100_000, exclude_uploaded=False):
        last_page = math.ceil(total_results / page_size)
        page_number = last_page - 1  # pages are 0 based
        while page_number >= 0:
            tmp = self.get_files_from_page(page_number, page_size, exclude_uploaded)
            if not tmp:
                break
            yield tmp
            page_number -= 1

    def batch_update_status(self, files, status=Status.COMPLETE):
        errors = []

        def to_payload(file):
            payload = {
                "id": file['id'],
                "md5sum": "None",
                "status": status,
                "size": file['size'] if 'size' in file else 0
            }
            return payload

        list_data = list(map(to_payload, files))
        url = "/".join([self.api, self.submission_id, 'files/batchUpdate'])
        data_to_dump = [list_data[i:i + self.batch_size] for i in range(0, len(list_data), self.batch_size)]
        for d in data_to_dump:
            data = json.dumps(d)
            response = put_request(url, payload=data, auth=self.auth)
            errors.extend(response['errors'])

        return errors

    @staticmethod
    def __upload_associated_file(local_associated_file, bucket, key, access_key, secret_key, session_token,
                                 transfer_config):
        file_name = str(local_associated_file.sys_path.resolve())
        s3 = get_s3_client_with_config(access_key,
                                       secret_key,
                                       session_token)
        s3.upload_file(file_name, bucket, key, Config=transfer_config)
        return local_associated_file

    def upload_associated_files(self, upload_progress=None, files_to_upload_count=None):
        batch_size = 50
        multipart_threshold = 5 * GB

        files_not_found = []
        file_not_found_count = 0
        errors_uploading_files = False

        if not files_to_upload_count:
            total_assoc_file_count, already_uploaded_count = self.get_upload_progress()
            files_to_upload_count = total_assoc_file_count - already_uploaded_count
        if not upload_progress:
            upload_progress = tqdm(total=files_to_upload_count, desc=f"Submission File Upload Progress",
                                   disable=self.config.hideProgress)

        transfer_config = TransferConfig(multipart_threshold=multipart_threshold, use_threads=False)
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.thread_num) as executor:
            for batch_number, submission_file_batch in enumerate(
                    self.get_files_iterator(files_to_upload_count, page_size=batch_size, exclude_uploaded=True)):
                # get credentials for this batch of files
                credentials_list = self.get_multipart_credentials(list([f['id'] for f in submission_file_batch]))
                id_to_credential = {int(cred['submissionFileId']): cred for cred in credentials_list}

                futures = []
                for file in submission_file_batch:
                    if not file['file_user_path'].startswith('s3://'):
                        local_file = to_local_associated_file(file, self.config.directory_list)
                        if not local_file:
                            if len(files_not_found) < 20:
                                files_not_found.append(file['file_user_path'])
                            file_not_found_count += 1
                            continue
                        else:
                            bucket, key = Utils.deconstruct_s3_url(file['file_remote_path'])
                            creds = id_to_credential[int(file['id'])]
                            access_key, secret_key, session_token = creds['access_key'], creds['secret_key'], creds[
                                'session_token']
                            futures.append(
                                executor.submit(self.__upload_associated_file, local_file, bucket, key, access_key,
                                                secret_key, session_token, transfer_config))
                    else:
                        raise Exception(f"Not implemented yet. File {file['file_user_path']} is in S3")
                completed_files_batch = []
                for future in concurrent.futures.as_completed(futures):
                    # noinspection PyBroadException
                    try:
                        local_associated_file = future.result()
                        completed_files_batch.append(local_associated_file.to_api_resource())
                        upload_progress.update(1)
                    except Exception:
                        errors_uploading_files = True
                        pass
                if completed_files_batch:
                    errors = self.batch_update_status(files=completed_files_batch)
                    if errors:
                        logger.warning(
                            f'Errors found when attempting to update status for batch# {batch_number}\n{errors}')

        if files_not_found or errors_uploading_files:
            if files_not_found:

                logger.error(f'Some files could not be found in {" ".join(self.config.directory_list)}:\n')
                for file in files_not_found:
                    logger.error(file)
                if file_not_found_count > 20:
                    logger.error(f'and {file_not_found_count - 20} more files.')
                self.config.directory_list = Utils.collect_directory_list()
                self.upload_associated_files(upload_progress, files_to_upload_count)
            else:
                logger.error(f'There were errors uploading files. \r\n'
                             f'Please try resuming the submission by running vtcmd -r {self.submission_id}\r\n'
                             f'If the error persists, contact NDAHelp@mail.nih.gov for help.')
                exit_error()
        else:
            upload_progress.close()

    def resume_submission(self):
        self.check_status()

        if self.status == Status.UPLOADING:
            self.upload_associated_files()
            self.check_status()
